
/*  -------------------------------
 * 
 * 
 *       TkApplication.cc
 * 
 * 
 */


#include <stdlib.h>
#include <iostream.h>
#include <tk.h>
#include <tclExtend.h>
#include "canvas.h"
#include "demo.h"



extern canvas * c;


/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */

int matherr() { cout << "Math Error" << endl; exit(EXIT_FAILURE); }
int *tclDummyMathPtr = (int *) matherr;


class GD_Exception
     {
     friend ostream& operator<< (ostream& os, const GD_Exception& g );
     private:
        const char *str;
     public:
        GD_Exception( const char* s): str(s) {}
     };
inline ostream& operator<< (ostream& os, const GD_Exception& g )
     { os << g.str; }

class ECO_exception 
    {
    friend ostream& operator<< ( ostream& os, const ECO_exception& e);
    private:
       const char* which; 
    public:
       static const char  Cannot_register []="Cannot_register" ;     
       static const char  Attempted_deallocation_of_extended_object []="Attempted_deallocation_of_extended_object" ; 
       static const char  Attempted_unextend_of_unregistred_extender []="Attempted_unextend_of_unregistred_extender" ;
       static const char  Internal_wrong_precompilde_code_non_extensible []= "Internal: wrong_precompilde_code_non_extensible";
       ECO_exception(const char * x): which(x) {};
    };

inline ostream& operator<< ( ostream& os, const ECO_exception& e)
    { os << e.which;  };


void error( const char* s )
    {
    cerr << s << endl;
    exit(EXIT_FAILURE);
    }




#define NewCommand( Name ) \
    Tcl_CreateCommand(interp, "GD_" #Name,    TclCommand_##Name ,   NULL, NULL);


    int
main(int argc, char ** argv)
    {
    Tcl_Interp* interp= Tcl_CreateInterp();
    if ( Tcl_Init(interp)  == TCL_ERROR) error( "Tcl_Init error" );
//    if ( Tclx_Init(interp) == TCL_ERROR) error( "Tclx_Init error" );
    if ( Tk_Init(interp)   == TCL_ERROR) error( "Tk_Init error" );
    
//    Tcl_StaticPackage(interp, "Tclx", Tclx_Init, Tclx_SafeInit);
    Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);
    
    // initialization of new commands

    NewCommand( Add_Vertex )  // modifing data
    NewCommand( Add_Edge )      
    NewCommand( Del_Vertex )
    NewCommand( Del_Edge )
    NewCommand( SetLabel )
    NewCommand( Show_Label )
    NewCommand( Layering )
      
    NewCommand( ConnectedInfo )    // info manager
    NewCommand( Oriented_Cycles_Info )
      
    NewCommand( Incidents )    // get data
    NewCommand( Extremes )
    NewCommand( State ) 
    NewCommand( CoordToVertex )

    NewCommand( Orientation )
    NewCommand( Polygonal )

    NewCommand( Unsplit_All )
      
      
    // GD initilization
    ofstream* log_str= NULL;
    if ( argc==2 && strcmp(argv[1], "-log")==0 ) 
        log_str= new ofstream("log.tcl");
    Create_GD(log_str);
    
    c= new canvas(interp, ".draw_frame.c" );
    
    
    // user interface initialization (tcl script)
    if ( Tcl_EvalFile( interp, "./Apprc.tcl" ) == TCL_ERROR ) error( "Tcl_EvalFile error" );

    try {
        Tk_MainLoop();
        }
    catch (char * s )
        {cerr << "(char*):" << s << endl;}
    catch (GD_Exception s )
        {cerr << "GD:" << s << endl;}
    catch (ECO_exception s )
        {cerr << "ECO:" << s << endl;}
    catch ( ... )
        { cerr << "Eccezione sconosciuta" << endl; }
    
    Destroy_GD();
    return 0;		     
    } 

