
#include <math.h>

#include <tcl.h>
#include <tk.h>

#include <iostream.h>
#include <strstream.h>


#include <canvas.h>

     void
canvas::
Eval(const char* c)
          {
	  char *s= new char[ strlen(c)+1 ];
	  strcpy(s, c);
	  int st= Tcl_Eval(interp, s);
          switch (st)
	      {
	      case TCL_OK: return; break;
	      case TCL_ERROR: s= "error"; break;
	      case TCL_RETURN: s= "return"; break;
	      case TCL_BREAK: s= "break"; break;
	      case TCL_CONTINUE: s= "continue"; break;
	      default: s="unknown" ; break;
	      }
	  cerr << "canvas::eval() ->Tcl error: "<< s << endl;
	  cerr << c;
	  cerr << endl;
	  }



canvas::
canvas(Tcl_Interp* i, const char * CanvasVariable):
          interp( i ),
          The_Canvas( CanvasVariable )
          {
	  }



        void
canvas::
DrawEdge(void* e,  unsigned from_x, unsigned from_y, unsigned to_x, unsigned to_y, Dir_Info dir )
          {
	  list<CoordPair> L;
	  CoordPair From; From.x= from_x; From.y=from_y;
	  CoordPair To;   To.x= to_x;     To.y=to_y;
	  L.push_front(From);
	  L.push_back(To);
	  DrawEdge(e, L, dir);
	  }


// Compute the correction to coordinate of an edge to not write edge on the 
// icon for the vertex (circular icon)
//
// Given two couple of coordinates (x,y) and (x1,y1)
//    it correct (x,y) computing (new_x,new_y)
//  (new_x,new_y) is the intersection point of a circle centered in 
//  (x,y) and a segment from (x,y) and (x1,y1)

          static void
circle_correction( double x,  double y,
                   double x1, double y1,
		   double r,
                   double& new_x, double& new_y
		  )               
          {
	  double d= sqrt( (x-x1)*(x-x1) + (y-y1)*(y-y1) );
	  double correction_x= (   d!=0   ?    (x1-x)* r / d   :   0   );
	  double correction_y= (   d!=0   ?    (y1-y)* r / d   :   0   );
	  new_x= x+ correction_x;
	  new_y= y+ correction_y;
	  }
		   
        void
canvas::
DrawEdge( void* e, const list<CoordPair>& C, Dir_Info dir )
          {
	  if ( C.size()<2 ) throw "canvas::DrawEdge() list of coordinates must have size >=2";
	                        
	                                     // first segement
          unsigned from_x= (*C.begin()).x;     // first point
          unsigned from_y= (*C.begin()).y;   
          unsigned from1_x= (*++C.begin()).x;  // second point
          unsigned from1_y= (*++C.begin()).y;
	  double f_x, f_y;  circle_correction(from_x,from_y, from1_x,from1_y, R_Icon, f_x,f_y );
	  
	                                     // last segment
          unsigned to1_x= (*-- --C.end()).x;  // the point before the last one
          unsigned to1_y= (*-- --C.end()).y;
          unsigned to_x= (*--C.end()).x;      // last point
          unsigned to_y= (*--C.end()).y;     
	  double t_x, t_y;  circle_correction(to_x,to_y, to1_x,to1_y, R_Icon, t_x,t_y );
	  
	    
	  ostrstream c;
	  c << The_Canvas << " create line "  << f_x << ' '  << f_y << ' ';
	  for( list<CoordPair>::const_iterator i= ++C.begin(); i!= --C.end(); ++i )
	        c << (*i).x << ' ' << (*i).y << ' ';
	  c << t_x << ' ' << t_y ;
	  
	  c.form(" -tags { edge r_e%p } ", e);
	  if ( dir!= undirected )
	     {
	     c << "-arrow " << ( dir==directed_forward ? "last" : "first" ) << ' ';
	     }
	  c <<" ;" << ends;
	  Eval(c.str());
	  delete c.str();
	  }





        void
canvas::
DrawVertex(void* v, unsigned x, unsigned y, const char *s )
          {
	  ostrstream c;
	  unsigned r= R_Icon;

	  c << The_Canvas << " create oval " 
	    << x-r << ' ' << y-r << ' '
	    << x+r << ' ' << y+r;
	  c.form( " -tags { vertex r_v%p } ", v );
	  c  <<" -fill yellow" << endl ;

          if (s)    // optional: label
	       {   
	       c << The_Canvas << " create text "
		 << x << ' ' << y;
	       c.form( " -tags { vertex r_t%p } ", v );
	       c << " -anchor center -text "
		 << s << endl ;
               }
	  
	  c << ends;
	  
	  Eval(c.str());
	  delete c.str();
	  }


/*
       void 
canvas::
DrawText(unsigned x, unsigned y, const char * str)
          {
	  ostrstream c;
	  c << The_Canvas << " create text -text " << '"' << str <<  '"' <<endl;
	  Eval(c.str());
	  delete c.str();
	  }
*/
       void 
canvas::
Clear()
          {
	  ostrstream c;
	  c << The_Canvas << " delete all" <<ends;
	  Eval(c.str());
	  delete c.str();
	  }
 
