#ifndef __WINDOW_ON_A_STREAM_H__
#define __WINDOW_ON_A_STREAM_H__

#include <iostream.h>

//Fifo: is a fixed capacity queue managed in fifo way
// T must have defined the following: T&  operator=( const T& )

template< class T, int l>
class Fifo
    {
    public:
       const int lenght=l;
    
       void put(T x)   // `x' is the first element, now. The last is dropped 
          {
	  position= (position==0) ? lenght-1 : position-1;
	  queue[position]= x;
	  }
       T& get(unsigned int w)   // access to each element of the fifo
          {
	  if ( w>=lenght ) throw "Fifo::get(w) --> `w' out of range";	    
	  return queue[(position+w) % lenght];
	  }
       Fifo():
           position(lenght-1)
           {}
       out(ostream& os) 
           {
	   for(int i=0; i<lenght ; i++)
	      os << get(i) << endl;
	   }
           
    private:
       T queue[lenght];
       int position;  // 0<= position <= lenght-1 ; this is the head of the fifo-queue, it's decremented and then wrapped  [see "T get(int)"]
    };





//////////////////////////////////////////////////////////////
//
// Window_on_a_Stream
//
// is a class that permit to manage a scrolling "window" on a stream
// (here we talk about generic stream not I/O stream)
// making available a certain amount of data before or after 
// a current position.
// It is useful when we don't have direct access to all element of the stream
// but just one at time.
//
//
// examples:
//
// this window has delta==7
// is possible to obtain (by get(i)) elements -3 through 3
// 
//
//    [ 6  5  4  3  2  1  0 --> numbering for fifo::get(i) ]
//
//     -3 -2 -1  0  1  2  3 
//      .  .  .  .  .  .  .  
//               |
//               |
//                ---> current position
//
// one can
//         get(0) to obtain the current element
//         get(-3) to obtain the previous -3th element
// one can't get(-4)
// but one can get(4) provoking a call to make_availible(4)
// one can't get(7) because 0 could be no more availible.
//
//
// after a shift_window()
// 
//  
//    [ 6  5  4  3  2  1  0 --> numbering for fifo::get(i) ]
//
//     -2 -1  0  1  2  3  4 
//      .  .  .  .  .  .  .  
//            |
//            |
//             ---> current position
//
// after a make_available(6) call shift_window() two times (until the requested element became available)
//

template< class T, int l, class M>
class Window_on_a_Stream
    {
    public:
       const delta=l;
    
       T& get(int i)
          {
	  make_available(i);
	  return f.get(position-i);
	  }
       T& get_and_advance()     // return the current element and advance it to the next one
          {
	  make_available(1);
	  position--;
	  return f.get(position+1); 
	  }
       Window_on_a_Stream(M& x, T (M::* GN)() ):
          position(0),
          server_class(x),
          GetNext(GN)
          {
	  get_and_advance(); // the first in the fifo is garbage
	  make_available(delta-1); // now one can get 0 trough delta-1, inhibit getting negative elements
	  }
       void out(ostream& os)
          {
	  for( int i=0; i<delta; i++ )
	     {
	     os << f.get(i);
	     if (i==position) os << "<---";
	     os << endl;
	     }
	  } 
    private:
       void shift_window()     // shift the window forward but the current element remain the same
          {
	  f.put( (server_class.*GetNext)() );
	  position++;
	  }
       void make_available(int i )  // shift window of a quantity necessari to make i avilable
          {
	  if ( i<=position && i>position-delta )     // already avilable?
	     ; // do nothing (already available)
	  else if ( i>position  // not avilable
		    && i< delta ) // shift admissible
	     while( position != i ) shift_window();
	  else if (i>=delta || i <= position-delta )
	     {
	     throw "Internal error: I can't make available.";
	     }
	  }
       Fifo<T,delta> f;
       T (M::* GetNext)();
       M& server_class;
    
                        // 0<= position <= delta-1
       int position;    // position of the fifo in which is the current element
                        // it correspond to the maximum index of available element
    };

#endif
