
// implementation of class Substituter
#include "backend.h"
#include "messages.h"


Substituter::
Substituter(istream& in, ostream& out):
    input(in),
    output(out)
    { };




    void 
Substituter::
write_seq( streampos h )
    {                          // write all till 'till_here'
    char* b= new char[h];
    input.read(b,h);
    if (input.eof())    M::hl_internal(M::Past_The_End);
    if (!input.good())  M::hl_error(M::Input_Error);
    
    
    
    output.write(b,h);
    if (!output)  M::hl_error(M::Output_Error);
    delete[] b;
    }


    void
Substituter::
freeze(streampos till_here)
    {
    if ( input.tellg()>till_here )
        {
	M::hl_internal(M::Useless_Freeze);
        return;
	}
    while(true)
        {
        if (changes.size()==0)
	    {                          // write all till 'till_here'
            write_seq( till_here - input.tellg() );
	    break;
	    }
        SSet::iterator i=changes.begin();
	const Substitution & g=*i;
	if ( g.where >= till_here )           // no substitution befor then
	    {                                    // write all till 'till_here'
            write_seq( till_here - input.tellg() );
	    break;
	    }
	else                  
	    {                          // there is a substitution before
            write_seq( g.where - input.tellg() );  // write till substitution
	    if ( g.where + g.how_many -1 >= till_here ) 
	        M::hl_internal(M::Freeze_in_Subst);
	    input.seekg(g.how_many, ios::cur);  // skip substituted part
 	    output << g.s;              // write substitution string
            if (!output)  M::hl_error(M::Output_Error);
	    delete g.s;    
	    changes.erase(i);
	    }

        }
    output << flush;
    };





    void
Substituter::
substitute( streampos where, streampos how_many, const char* s )
    {
                            // check preconditions
    if ( !s )                   M::hl_internal(M::Subst_NULL);
    if ( where< input.tellg() ) M::hl_internal(M::Already_Freezed);
    
                            // the real body
         {
         Substitution x;
         x.where   =where;
         x.how_many=how_many;
	 x.s       =s;
         pair<SSet::iterator, bool> p=changes.insert(x);
	 if (!p.second) M::hl_internal(M::Subst_Exist);
	 
	 
	 SSet::iterator next=p.first; next++;
	 SSet::iterator prec=p.first; prec--;
	 if ( p.first!=changes.begin() )  // if it is not the first
	      {
              const Substitution& y= *prec;
	      if ( y.where+y.how_many-1 >= x.where )  M::hl_internal(M::Subst_Overlap);
	      };
	 if ( next!=changes.end() )  // if it is not the last
	      {
              const Substitution& y= *next;
	      if ( x.where+x.how_many-1 >= y.where ) M::hl_internal(M::Subst_Overlap);
	      };
	 
         }
    };
#undef s_internal




    void
Substituter::
output_changes() const
    {
    for( SSet::const_iterator k=changes.begin(); k!= changes.end(); k++ )
         {
	 cerr
	      << "-------- " << (*k).where << ' '  << (*k).how_many << endl 
	      << (*k).s << endl;
	 }
    };




//==============================================
/*           this could be a test

#include <fstream.h>
#include <strstream.h>
int main()
    {
    ifstream in("xyz");
    Substituter s(in, cout);
    strstream m,n,p;
    m << "<1>" << ends;;
    s.substitute( 5, 3, m.str() );
    n << "<2>" << ends;
    s.substitute( 8, 3, n.str() );
    s.freeze(12);
    p << "<3>" << ends;
    s.substitute( 18, 3, p.str() );
    s.freeze( 22 );
    return 0;
    };
 */
