
#ifndef __BACKEND_H__
#define __BACKEND_H__

// This is the module that perform writing on an output stream what is read in from an
// input stream but with the possibility to substitute portion of text.
// Delete part of the input and insertion without substitution are also possible as
// special cases.

#include <stdlib.h>
#include <iostream.h>
#include <set.h>

class Substituter
     {
     private:
         struct Substitution
            {
	    streampos where;
	    streampos how_many;
	    const char* s;
	    };
         struct SubstitutionLess
            {
	    bool operator() (const Substitution& a, const Substitution& b) const
	        {return a.where<b.where;}
	    };
         typedef set< Substitution, SubstitutionLess > SSet;
     
     
         istream& input;
         ostream& output;
     
         SSet changes;    // contain changes to input stream
     
         void write_seq( streampos x );   // copy on 'output' 'x' character form 'input'
     public:
         Substituter(istream& in, ostream& out);
              // WARNING: streams specified by 'in' and 'out' are heavily managed by
              //    Substituter (mainly changing write an read position).
              //    Don't use it after creting an instance of Substituter for them.
         
         void freeze(streampos till_here); 
              // It provoke writing on 'output' the text corresponding to 'input' till
              // the position pointed by 'till_here' (not included), but with the changes
              // specified by calling substiute(...).
              // The part of input from start to input.tellg()-1 is called "freezed area"
          
         void substitute( streampos where, streampos how_many, const char* s );
              // preconditions: 
              //     s!= NULL; 
              //     where, where+how_many valid position for input
              //     where >= input.tellg()     // it must not change the freezed area
              //     no other substitutions overlap this one
              //
              // the NULL terminated string pointed by 's' became delete-owned by this instance of Substituter
              // 
              //
              // special cases:
              //  delete --> s="", deletion begin in where for how_many characters
              //  insert --> s+ "string to insert", how_many==0; insertion happen in (before) where
         
         // debug
         void output_changes() const;
     };


#endif
