
// implementation of class Substituter
#include "backend.h"



Substituter::
Substituter(istream& in, ostream& out):
    input(in),
    output(out)
    { };

    void
Substituter::
internal(const char *s) const
    {
    cerr << "internal error in class Substituter (the backend):" << endl;
    cerr << s << endl;
    exit(EXIT_FAILURE);
    };



    void 
Substituter::
write_seq( streampos h )
    {                          // write all till 'till_here'
    char* b= new char[h];
    input.read(b,h);
    if (!input.good())
        internal("Substituter::write_seq(...): read past the end of input or other problem\n"
		 "A read error is also possible"
		 );
        
    output.write(b,h);
    if (!output)
        {
        cerr << "problem writing output." << endl;
        exit(EXIT_FAILURE);
        }
    delete[] b;
    }


    void
Substituter::
freeze(streampos till_here)
    {
    if ( input.tellg()>till_here )
        {
	cerr << "waring: Substituter::freeze(streampos till_here), useless freeze("
	     << till_here << ')' << endl;
        return;
	}
    while(true)
        {
        if (changes.size()==0)
	    {                          // write all till 'till_here'
            write_seq( till_here - input.tellg() );
	    break;
	    }
        SSet::iterator i=changes.begin();
	const Substitution & g=*i;
	if ( g.where >= till_here )           // no substitution befor then
	    {                                    // write all till 'till_here'
            write_seq( till_here - input.tellg() );
	    break;
	    }
	else                  
	    {                          // there is a substitution before
            write_seq( g.where - input.tellg() );  // write till substitution
	    if ( g.where + g.how_many -1 >= till_here )
	        internal("freeze(...): freeze in the middle of a substitution.");
	    input.seekg(g.how_many, ios::cur);  // skip substituted part
	    output << g.s;              // write substitution string
            if (!output)
                {
                cerr << "problem writing output." << endl;;
                exit(EXIT_FAILURE);
                }
	    delete g.s;    
	    changes.erase(i);
	    }

        }
    output << flush;
    };




// this is a local macro
// auto string concatenation of C is used 
#define s_internal(s) \
      {\
      cerr << "================ error" << endl; \
      output_changes();    \
      internal("in calling substiute(...): " s );\
      }

    void
Substituter::
substitute( streampos where, streampos how_many, const char* s )
    {
                            // check preconditions
    if ( !s )                   s_internal("s==NULL");
    if ( where< input.tellg() ) s_internal("where< input.tellg()");
    
    /*  
     these doesn't detect the errors he wont to detect. It must effectively do the read to produce an error.
     I don't wont to do this for efficency reason.
     
    streampos old_pos= input.tellg();
    input.seekg(where);          if( !input.good() ) s_internal("'where' is not valid");
    input.seekg(where+how_many); if( !input.good() ) s_internal("'where+how_many' is not valid");
    input.seekg(old_pos);        if( !input.good() ) s_internal("unknowun problem");
    */
    
                            // the real body
         {
         Substitution x;
         x.where   =where;
         x.how_many=how_many;
	 x.s       =s;
         pair<SSet::iterator, bool> p=changes.insert(x);
	 if (!p.second) s_internal("a change for this value of 'where' already exist");
	 
	 
	 SSet::iterator next=p.first; next++;
	 SSet::iterator prec=p.first; prec--;
	 if ( p.first!=changes.begin() )  // if it is not the first
	      {
              const Substitution& y= *prec;
	      if ( y.where+y.how_many-1 >= x.where ) s_internal("sovrapposition whith precedent");
	      };
	 if ( next!=changes.end() )  // if it is not the last
	      {
              const Substitution& y= *next;
	      if ( x.where+x.how_many-1 >= y.where ) s_internal("sovrapposition whith next");
	      };
	 
         }
    };
#undef s_internal




    void
Substituter::
output_changes() const
    {
    for( SSet::const_iterator k=changes.begin(); k!= changes.end(); k++ )
         {
	 cerr
	      << "-------- " << (*k).where << ' '  << (*k).how_many << endl 
	      << (*k).s << endl;
	 }
    };




//==============================================


#include <fstream.h>
#include <strstream.h>
int main()
    {
    ifstream in("xyz");
    ifstream in2("xyz");
    char x;
    in2 >> x;
    cout << "read from in2: " << x;

    Substituter s(in, cout);
    strstream m,n,p;
    m << "<1>" << ends;;
    s.substitute( 5, 3, m.str() );
    n << "<2>" << ends;
    s.substitute( 8, 3, n.str() );
    s.freeze(12);
    p << "<3>" << ends;
    s.substitute( 18, 3, p.str() );
    s.freeze( 21 );
    return 0;
    };

