

 /* -----------------------------------------------------------
  * cxxlex.l: modulo di analisi lessicale per il precompilatore
  *           Scritto da Maurizio Pizzonia
  *
  *  $Id: $
  */

/* yywrap  una funzione che dovrebbe servire per gestire meccanismo in stile #include */
%option noyywrap
/* yylineno e` una variabile che contiene il numero di linea corrente */
%option yylineno
                     /* ti da l'opportunita' di attivare il debug */
%option debug 
                     /* cosi' lo scanner diventa una classe */
%option c++   
%option yyclass="Cxx_lexer"

%{
#include <assert.h>
#include <stdlib.h>
#include <iostream.h>
//#include <string.h>

#include "frontend.h"   // definizione di parser e token e inclusione dei simboli generati da BISON

#include "pardefs.h"


// questo cambia il nome della funzione nella classe (solo per C++)
#undef  YY_DECL
#define YY_DECL A_token Cxx_lexer::inner_lexer()



// some macro for standard procedure.




#define Update_and_emit(t) \
       {                   \
       A_token token;      \
       almost_compute_token( token, t ); \
       return token;        \
       }



#define Update_and_emit_aux(t, kind, value)   \
       {                                \
       A_token token;                    \
       almost_compute_token( token, t ); \
       token->aux.kind= value;           \
       return token;                     \
       }


%}

Separatore        {Separatore_no_NL}|"\n"
Separatore_no_NL  " "|"\t"|"\r"|"\f"
				/* altre abbreviazioni */
Cifra             [0-9]
Alfa              [A-Za-z_]
AlfaNumerico      {Cifra}|{Alfa}
Numero_Intero1    {Cifra}({Cifra}*)
Numero_Intero     {Numero_Intero1}("L"?)
Identificatore    {Alfa}({AlfaNumerico}*)
Internal_of_String      (\\\")|[^\"\n]
Internal_of_Char_const  (\\\')|[^\'\n]

Stringa           \"({Internal_of_String}*)\"
Char_const        \'({Internal_of_Char_const}*)\'
Numero_FloatingPoint   ({Numero_Intero}".")|("."{Numero_Intero})|({Numero_Intero}"."{Numero_Intero})

PreProcessor_Info_Line "#"(.*)"\n"
PreProcessor_Pragma_Line "#pragma"(.*)"\n"

%%
{Separatore_no_NL}              { Update_and_not_emit(); }
"\n"                       { current_line++; Update_and_not_emit(); }
{PreProcessor_Pragma_Line} { current_line++; Update_and_not_emit(); }
{PreProcessor_Info_Line}  {
                          char* buff= new char[strlen(YYText()+1)+1];
                          strcpy(buff, YYText()+1);    // copy stripping '#'
                          char* n= strchr( buff, '"');
                          if ( n ) 
                             {
                             *n='\0';
                             current_line= atoi(buff);   // this is the 'current_line' value for the next line (see CPP info pages)
                             char* m= strchr( n+1, '"' );
			     if (m)
			       {
                               *m='\0';
                               current_include=files_included.Update(n+1);
			       }
			     }
                          delete buff;
                          Update_and_not_emit(); 
                          }


  /* these are taken from ANSI working paper, those unimplemented are comented */

"auto"              { Update_and_emit_aux(SCSPEC, sc, Auto);}
"bool"              { Update_and_emit_aux(TYPESPEC, ts,  Bool);}
"break"             { Update_and_emit(BREAK);}
"case"              { Update_and_emit(CASE);}
"catch"             { Update_and_emit(CATCH);}
"char"              { Update_and_emit_aux(TYPESPEC, ts, Char);}
"class"             { Update_and_emit(AGGR);}
"const"             { Update_and_emit_aux(TYPE_QUAL, tq, Const);}
 /*"const_cast"*/
"continue"          { Update_and_emit(CONTINUE);}
"default"           { Update_and_emit(DEFAULT);}
"delete"            { Update_and_emit(DELETE);}
"do"                { Update_and_emit(DO);}
"double"            { Update_and_emit_aux(TYPESPEC, ts, Double);}
"dynamic_cast"      
"else"              { Update_and_emit(ELSE);}
"enum"              { Update_and_emit(ENUM);}
  /*"explicit"*/     
"extern"            { Update_and_emit_aux(SCSPEC, sc, Extern);}   
"false"             { Update_and_emit(CXX_FALSE);}
"float"             { Update_and_emit_aux(TYPESPEC, ts, Float);}
"for"               { Update_and_emit(FOR);}
"friend"            { Update_and_emit_aux(SCSPEC, sc, Friend);}
"goto"              { Update_and_emit(GOTO);}
"if"                { Update_and_emit(IF);}
"inline"            { Update_and_emit_aux(SCSPEC, sc, Inline);}
"int"               { Update_and_emit_aux(TYPESPEC, ts, Int);}
"long"              { Update_and_emit_aux(TYPESPEC, ts, Long);}
"mutable"           { Update_and_emit_aux(SCSPEC, sc, Mutable);}
  /*"namespace"*/        
"new"                { Update_and_emit(NEW);}
"operator"           { Update_and_emit(OPERATOR);}
"private"            { Update_and_emit_aux(VISSPEC, vs, Private);}
"protected"          { Update_and_emit_aux(VISSPEC, vs, Protected);}
"public"             { Update_and_emit_aux(VISSPEC, vs, Public);}
"register"           { Update_and_emit_aux(SCSPEC, sc, Register);}
  /*"reinterpret_cast"*/ 
"return"             { Update_and_emit(RETURN);}
"short"              { Update_and_emit_aux(TYPESPEC, ts, Short);}
"signed"             { Update_and_emit_aux(TYPESPEC, ts, Signed);}
"sizeof"             { Update_and_emit(SIZEOF);}
"static"             { Update_and_emit_aux(SCSPEC, sc, Static);}
  /* "static_cast" */  
"struct"             { Update_and_emit(AGGR);}
"switch"             { Update_and_emit(SWITCH);}
"template"           { Update_and_emit(TEMPLATE);}
"this"               { Update_and_emit(THIS);}
"throw"              { Update_and_emit(THROW);}
"true"               { Update_and_emit(CXX_TRUE);}
"try"                { Update_and_emit(TRY);}
"typedef"            { Update_and_emit_aux(SCSPEC, sc, Typedef);}
  /* "typeid"   */
  /* "typename" */
"union"              { Update_and_emit(AGGR);}
"unsigned"           { Update_and_emit_aux(TYPESPEC, ts, Unsigned);}
  /* "using"    */
"virtual"            { Update_and_emit_aux(SCSPEC, sc, Virtual);}
"void"               { Update_and_emit_aux(TYPESPEC, ts, Void);}
"volatile"           { Update_and_emit_aux(TYPE_QUAL, tq, Volatile);}
"wchar_t"            { Update_and_emit_aux(TYPESPEC, ts, Wchar_t);}
"while"              { Update_and_emit(WHILE);}



    /* G++ extensions */
"__attribute"|"__attribute__"     { Update_and_emit(ATTRIBUTE); }
"__alignof__"                     { Update_and_emit(ALIGNOF); }
"__extension__"                   { Update_and_emit(EXTENSION); }
"__const"                         { Update_and_emit_aux(TYPE_QUAL, tq, Const);}
"__signed__"|"__signed"           { Update_and_emit_aux(TYPESPEC, ts, Signed);}
    /* end G++ extensions */




   /* ECO extensions */
"___ECO_C++_sign___"              { Update_and_emit(ECO_CXX_SIGN); }

".{"|("."{Separatore}"{")         { Update_and_emit(POINT_LEFTBRACE); }
"}."|("}"{Separatore}".")         { Update_and_emit(RIGHTBRACE_POINT); }

"-{"|("-"{Separatore}"{")         { Update_and_emit(MINUS_LEFTBRACE); }
"}>"|("}"{Separatore}">")         { Update_and_emit(RIGHTBRACE_GREATER); }

"extensible"                      { Update_and_emit(EXTENSIBLE); }
"extend"                          { Update_and_emit(EXTEND); }
"dynamic"                         { Update_and_emit(DYNAMIC); }
   /* end ECO extensions */






"{"|"}"|"["|"]"|"="|"("|")"|"~"|"."|";"|":"|"?"|","|"+"|"-"|"*"|"/"|"%"|"^"|"&"|"|"|"!"|"="|"<"|">"   {Update_and_emit(*YYText());}



{Numero_Intero}|{Numero_FloatingPoint}|{Char_const}   { Update_and_emit(CONSTANT); }
{Stringa}                      { Update_and_emit(STRING); }
{Identificatore}               {
                               const char *s;
                               current_scope.Symbol_LookUp(YYText(), s);
                               value_identifier a_id;
                               a_id.s= s;
                               a_id.e= NULL;
                               Update_and_emit_aux(IDENTIFIER, id, a_id);
                               }


"+="  { Update_and_emit_aux(ASSIGN, op, Sum_Assign); }
"-="  { Update_and_emit_aux(ASSIGN, op, Sub_Assign); }
"*="  { Update_and_emit_aux(ASSIGN, op, Mul_Assign); }
"/="  { Update_and_emit_aux(ASSIGN, op, Div_Assign); }
"%="  { Update_and_emit_aux(ASSIGN, op, Mod_Assign); }
"^="  { Update_and_emit_aux(ASSIGN, op, Bit_Xor_Assign); }
"&="  { Update_and_emit_aux(ASSIGN, op, Bit_And_Assign); }
"|="  { Update_and_emit_aux(ASSIGN, op, Bit_Or_Assign); }
">>=" { Update_and_emit_aux(ASSIGN, op, RShift_Assign); }
"<<=" { Update_and_emit_aux(ASSIGN, op, LShift_Assign); }










"++"       { Update_and_emit(PLUSPLUS); }
"--"       { Update_and_emit(MINUSMINUS); }

"=="  { Update_and_emit_aux(EQCOMPARE, op, Equal); }
"!="  { Update_and_emit_aux(EQCOMPARE, op, Not_Equal); }

"<="  { Update_and_emit_aux(ARITHCOMPARE, op, Less_Equal); }
">="  { Update_and_emit_aux(ARITHCOMPARE, op, Greater_Equal); }

">>"       { Update_and_emit(RSHIFT); }
"<<"       { Update_and_emit(LSHIFT); }

"||"       { Update_and_emit(OROR); }
"&&"       { Update_and_emit(ANDAND); }
 
"::"       { Update_and_emit(SCOPE); }

"("{Separatore}*")"                       { Update_and_emit(LEFT_RIGHT); /* WARNING: POSSIBLE ERROR ON LINE NUMBER*/}
"("{Separatore}*"*"{Separatore}*")"       { Update_and_emit(PAREN_STAR_PAREN); /* WARNING: POSSIBLE ERROR ON LINE NUMBER*/}

"..."             { Update_and_emit(ELLIPSIS); }
"->"              { Update_and_emit(POINTSAT); }
"->*"             { Update_and_emit(POINTSAT_STAR); }
".*"              { Update_and_emit(DOT_STAR); }

<<EOF>>   {  // almost like Update_and_emit(END_OF_COMPILATION_UNIT); but we don't wont to include EOF in the string
          A_token token;                                
          token->begin_pre_string = beginning_of_pre_string;
          token->begin_token_string= -1;               // no meaning
          Update_stream_position();                       
          token->end_token_string = stream_position-1;  // don't include EOF
          beginning_of_pre_string = stream_position;    
          token->tokenid=END_OF_COMPILATION_UNIT;                             
          return token;                                 
          }








.                { Update_and_not_emit(); cerr << "`"<< *YYText() << "' ignored. " << endl;  }



%%

     void
Cxx_lexer::
Update_stream_position()
     {
     stream_position += YYLeng();
     }



       void
Cxx_lexer::
Update_and_not_emit()
       {
       if (last_emitted) beginning_of_pre_string= stream_position; 
       last_emitted=false;     
       Update_stream_position();
       }
 
       void
Cxx_lexer::
almost_compute_token( A_token& token, t_id t)
       {                   
       token->begin_pre_string  = beginning_of_pre_string; 
       token->begin_token_string = stream_position;   
       Update_stream_position();                     
       token->end_token_string  = stream_position-1; 
       beginning_of_pre_string  = stream_position;   
       token->line= lineno();
       token->include_file= current_include;
       token->include_file_line= current_line;
       token->tokenid=t;                             
       }

Symbol_set Cxx_lexer::files_included;

