

/* This grammar is based on the GNU g++ 2.7.2 grammar.  */

/* Note: Bison automatically applies a default action of "$$ = $1" for
   all derivations; this is applied before the explicit action, if one
   is given.  Keep this in mind when reading the actions.  */

/* this messge is normal, the confilicts derive from GNU g++ 2.7.2 grammar */
/* cxxparsr.y contains 5 shift/reduce conflicts and 38 reduce/reduce conflicts. */

%pure_parser
%start compilation_unit
/* %start program */

%{
#include <strstream.h>

class token;
class Entity;
class AggregateType_Entity;
class Function;


#include "entity.h"
#include "parser_types.h"
%}

%union          // in the following names U mean "union"
   { 
   
   A_identifier            UA_identifier;    // special types created for parser
   A_declmods              UA_declmods ;
   A_typed_declspecs       UA_typed_declspecs;
   A_typespecqual_reserved UA_typespecqual_reserved;
   A_name_entity           UA_name_entity;
   A_name_entity_type      UA_name_entity_type;
   A_name_entity_scope_type UA_name_entity_scope_type;
   A_name_entity_scope     UA_name_entity_scope;
   A_Entity_endstring      UA_Entity_endstring;
   A_Entity_Name_endstring UA_Entity_Name_endstring;
   A_expression            UA_expression;
   A_object                UA_object;    // used only for non terminal "object"
   A_EntityContainer       UA_EntityContainer;
   A_type                  UA_type;
   A_Operator_Identifier   UA_Operator_Identifier;
     
   A_token                 UA_token;           // tokens


   Type*                   U_Type;           // type already existent
   Function*               U_Function;
   Entity*                 U_Entity;
   AggregateType_Entity*   U_AggregateType_Entity;
   Entity_Name*            U_Entity_Name;
   TYPE_QUAL_value         U_TYPE_QUAL_value;
   Grouping_Entity*        U_Grouping_Entity;
   Entity_Container*       U_Entity_Container;
   Template_Entity*        U_Template_Entity;
   list<Type*>*            U_ListTypes;
   streampos               U_streampos;
   common                  U_common;
   }; 

/* All identifiers that are not reserved words
   and are not declared typedefs in the current block */
%token IDENTIFIER


/* All identifiers that are declared typedefs in the current block.
   In some contexts, they are treated just like IDENTIFIER,
   but they can also serve as typespecs in declarations.  */
%token TYPENAME


/* Reserved words that specify storage class.
   yylval contains an IDENTIFIER_NODE which indicates which one.  */
%token SCSPEC

/* Reserved words that specify type.
   yylval contains an IDENTIFIER_NODE which indicates which one.  */
%token TYPESPEC

/* Reserved words that qualify type: "const" or "volatile".
   yylval contains an IDENTIFIER_NODE which indicates which one.  */
%token TYPE_QUAL

/* Character or numeric constants.
   yylval is the node for the constant.  */
%token CONSTANT

/* String constants in raw form.
   yylval is a STRING_CST node.  */
%token STRING

/* "...", used for functions with variable arglists.  */
%token ELLIPSIS

/* the reserved words */
/* SCO include files test "ASM", so use something else. */
%token SIZEOF ENUM /* STRUCT UNION */ IF ELSE WHILE DO FOR SWITCH CASE DEFAULT
%token BREAK CONTINUE RETURN GOTO ASM_KEYWORD GCC_ASM_KEYWORD TYPEOF ALIGNOF
%token SIGOF
%token ATTRIBUTE EXTENSION LABEL

/* the reserved words... C++ extensions */
%token AGGR
%token VISSPEC
%token DELETE NEW OVERLOAD THIS OPERATOR CXX_TRUE CXX_FALSE
%token NAMESPACE TYPENAME_KEYWORD USING
%token LEFT_RIGHT TEMPLATE
%token TYPEID DYNAMIC_CAST STATIC_CAST REINTERPRET_CAST CONST_CAST
%token  SCOPE

/* Define the operator tokens and their precedences.
   The value is an integer because, if used, it is the tree code
   to use in the expression made from the operator.  */

%left EMPTY                     /* used to resolve s/r with epsilon */

%left error

/* Add precedence rules to solve dangling else s/r conflict */
%nonassoc IF
%nonassoc ELSE

%left IDENTIFIER TYPENAME PTYPENAME SCSPEC TYPESPEC TYPE_QUAL ENUM AGGR ELLIPSIS TYPEOF SIGOF OPERATOR NSNAME TYPENAME_KEYWORD

%left '{' ',' ';'

%nonassoc THROW
%right  ':'
%right  ASSIGN '='
%right  '?'
%left  OROR
%left  ANDAND
%left  '|'
%left  '^'
%left  '&'
%left  MIN_MAX
%left  EQCOMPARE
%left  ARITHCOMPARE '<' '>'
%left  LSHIFT RSHIFT
%left  '+' '-'
%left  '*' '/' '%'
%left  POINTSAT_STAR DOT_STAR
%right  UNARY PLUSPLUS MINUSMINUS '~'
%left HYPERUNARY
%left PAREN_STAR_PAREN LEFT_RIGHT
%left  POINTSAT '.' '(' '['

%right SCOPE                    /* C++ extension */
%nonassoc NEW DELETE TRY CATCH

/* %type <code> unop */

/* %type <ttype> CONSTANT expr nonnull_exprlist */
/* %type <ttype> paren_expr_or_null nontrivial_exprlist */
/* %type <ttype> expr_no_commas cast_expr unary_expr primary string STRING */
/* %type <ttype> reserved_declspecs boolean.literal */
/* %type <ttype> reserved_typespecquals */
/* %type <ttype> nonempty_type_quals maybe_type_qual */
/* %type <itype> initdecls notype_initdecls initdcl      C++ modification  */
/* %type <ttype> init initlist maybeasm maybe_init */
/* %type <ttype> asm_operands nonnull_asm_operands asm_operand asm_clobbers */
/* %type <ttype> maybe_attribute attributes attribute attribute_list attrib */
/* %type <ttype> any_word */

/* %type <ttype> compstmt implicitly_scoped_stmt */

/* %type <ttype> after_type_declarator */
/* %type <ttype> direct_after_type_declarator */

/* %type <ttype> opt.component_decl_list component_decl_list */
/* %type <ttype> component_decl component_decl_1 components notype_components */
/* %type <ttype> component_declarator component_declarator0 */
/* %type <ttype> notype_component_declarator notype_component_declarator0 */
/* %type <ttype> after_type_component_declarator after_type_component_declarator0 */
/* %type <ttype> enumlist enumerator */
/* %type <ttype> type_id absdcl type_quals */
/* %type <ttype> direct_abstract_declarator conversion_declarator */
/* %type <ttype> new_type_id new_declarator direct_new_declarator */
/* %type <ttype> xexpr parmlist parms parm bad_parm full_parm */
/* %type <ttype> identifiers_or_typenames */
/* %type <ttype> fcast_or_absdcl regcast_or_absdcl */
/* %type <ttype> expr_or_declarator complex_notype_declarator */
/* %type <ttype> unqualified_id qualified_id */
/* %type <ttype> overqualified_id notype_qualified_id any_id */
/* %type <ttype> complex_direct_notype_declarator functional_cast */
/* %type <ttype> named_parm complex_parmlist parms_comma */

/* C++ extensions */
%token TYPENAME_ELLIPSIS PTYPENAME 
%token PRE_PARSED_FUNCTION_DECL EXTERN_LANG_STRING ALL 
%token PRE_PARSED_CLASS_DECL 
/* %type <ttype> fn.def1 /* Not really! */
/* %type <ttype> fn.def2 return_id */
/* %type <itype> ctor_initializer_opt */
/* %type <ttype> */
/* %type <ttype> named_complex_class_head_sans_basetype */
/* %type <ttype> */
/* %type <ttype> base_class_list */
/* %type <itype> base_class_access_list */
/* %type <ttype> base_class maybe_base_class_list base_class.1 */
/* %type <ttype> exception_specification_opt ansi_raise_identifier ansi_raise_identifiers */
/* %type <ttype> operator_name */
/* %type <ttype> object aggr */
/* %type <itype> new delete */
/* %type <ttype> primary_no_id */
/* %type <ttype> nonmomentary_expr maybe_parmlist */
/* %type <itype> initdcl0 notype_initdcl0 member_init_list */
/* %type <ttype> template_header template_parm_list template_parm */
/* %type <ttype> template_type_parm */
/* %type <ttype> template_type template_arg_list template_arg */
/* %type <ttype> template_instantiation template_type_name tmpl.2 */
/* %type <ttype> template_instantiate_once template_instantiate_some */
/* %type <itype> fn_tmpl_end */
/* %type <itype> try_for_typename */
/* %type <ttype> condition xcond paren_cond_or_null */
/* %type <ttype> type_name nested_name_specifier nested_type ptr_to_mem */
/* %type <ttype> qualified_type_name complete_type_name */
/* %type <ttype> complex_type_name nested_name_specifier_1 */
/* %type <itype> nomods_initdecls nomods_initdcl0 */
/* %type <ttype> new_initializer new_placement specialization type_specifier_seq */
/* %type <ttype> using_decl .poplevel */

/* in order to recognize aggr tags as defining and thus shadowing. */
%token TYPENAME_DEFN IDENTIFIER_DEFN PTYPENAME_DEFN
/* %type <ttype> PTYPENAME_DEFN */

%token NSNAME
/* %type <ttype> NSNAME */

/* Used in lex.c for parsing pragmas.  */
%token END_OF_LINE

/* lex.c and pt.c depends on this being the last token.  Define
   any new tokens before this one!  */
%token END_OF_SAVED_INPUT


//////////////////////////////////////////////////////////////////// The following is created by Maurizio

                  /* just a few tokens added respect g++2.7.2 grammar */
%token            END_PRE_PARSED

%token<UA_token>  END_OF_COMPILATION_UNIT   
%token<UA_token>  EXTENSIBLE EXTEND DYNAMIC
%token<UA_token>  POINT_LEFTBRACE RIGHTBRACE_POINT MINUS_LEFTBRACE RIGHTBRACE_GREATER
%token<UA_token>  ECO_CXX_SIGN

%token<UA_token>  PSEUDO_TYPE
                  /* end of added tokens */

%type <UA_token> NSNAME PTYPENAME_DEFN 
%type <UA_token>   SCSPEC
%type <UA_token>   VISSPEC
%type <UA_token>   TYPESPEC
%type <UA_token>   TYPE_QUAL
%type <UA_token>   IDENTIFIER IDENTIFIER_DEFN 
%type <UA_token>   TYPENAME TYPENAME_DEFN
%type <UA_token>   PTYPENAME 

%type <UA_token>   ';' ASSIGN EQCOMPARE ARITHCOMPARE  EXTENSION ANDAND SIZEOF ALIGNOF NEW DELETE

%type <UA_token>   '(' ')' '{' '}'  STRING ':' '=' RETURN '>' '-' '+' PLUSPLUS MINUSMINUS '!'
%type <UA_token>   '*' '&' '~' CONSTANT LEFT_RIGHT '[' ']'  THIS '/' '%' '|' '^' ',' '<'
%type <UA_token>   DYNAMIC_CAST CONST_CAST STATIC_CAST REINTERPRET_CAST  LSHIFT RSHIFT 
%type <UA_token>   TYPEID  '.' POINTSAT CXX_TRUE CXX_FALSE SCOPE OPERATOR OROR  MIN_MAX
%type <UA_token>   POINTSAT_STAR

 
/*------------------------------*/


  
  

%type <U_TYPE_QUAL_value>        nonempty_type_quals
%type <U_TYPE_QUAL_value>        type_quals

%type <U_Type>                 type_id 

    /* parmlist */
%type <U_Type>                 maybe_parmlist
%type <U_Function>             begin_parmlist
  
  

%type <UA_Entity_endstring>    complete_type_name
%type <UA_Entity_endstring>    qualified_type_name
%type <UA_Entity_endstring>    type_name
%type <U_Entity>               complex_type_name
%type <UA_Entity_endstring>    nested_type


%type <U_AggregateType_Entity> class_head
%type <U_AggregateType_Entity> named_class_head_sans_basetype
%type <U_AggregateType_Entity> named_class_head_sans_basetype_defn
%type <U_Grouping_Entity> structsp
%type <U_AggregateType_Entity> unnamed_class_head named_class_head 


%type <UA_Operator_Identifier>  operator_name 

/* types declaration sub-grammars */
        /* component_declarator */
%type <UA_name_entity_type> after_type_component_declarator0
%type <UA_name_entity_type> after_type_component_declarator
%type <UA_name_entity_type> notype_component_declarator 
%type <UA_name_entity_type> notype_component_declarator0
        /* declarator */
%type <UA_name_entity_scope_type> after_type_declarator
%type <UA_name_entity_scope_type> direct_after_type_declarator
%type <UA_name_entity_type> expr_or_declarator
%type <UA_name_entity_scope_type> direct_notype_declarator
%type <UA_name_entity_scope_type> complex_notype_declarator 
%type <UA_name_entity_scope_type> complex_direct_notype_declarator
%type <UA_name_entity_scope_type> notype_declarator
        /* abstract */
%type <U_Type>                 absdcl 
%type <U_Type>                 conversion_declarator
%type <U_Type>                 direct_abstract_declarator
        /* id */
%type <UA_name_entity_scope>      notype_qualified_id
%type <UA_name_entity_scope>      qualified_id
%type <UA_name_entity_scope>      overqualified_id
%type <UA_name_entity_scope>      any_id
%type <UA_name_entity>      notype_unqualified_id
%type <UA_name_entity>      unqualified_id
       /* templates */
%type <U_Type>                 template_arg
%type <U_ListTypes>            template_arg_list
%type <UA_Entity_Name_endstring> template_type_name
%type <UA_Entity_endstring>      template_type
%type <U_Template_Entity>     template_header

       /* exprssions */
%type <UA_expression>          primary unary_expr cast_expr expr_no_commas expr 
%type <UA_expression>          functional_cast
%type <U_ListTypes>            nonnull_exprlist nontrivial_exprlist
%type <UA_token>               unop
%type <UA_type>                 regcast_or_absdcl
%type <UA_type>                 new_type_id
%type <UA_object>              object
%type <U_Type>                 new_declarator direct_new_declarator
       /* other */
%type <UA_EntityContainer>    nested_name_specifier_1
%type <UA_identifier>  identifier identifier_defn notype_identifier
%type <UA_token>       skip_till_right_brace
%type <UA_token>       skip_till_left_brace
%type <UA_token>       left_curly
%type <U_Entity>       fn.def2 declarator fn.def1
%type <U_common>       boolean.literal string new_initializer global_scope fcast_or_absdcl init
%type <U_common>       nested_name_specifier new delete operator
 

%{
//#include <string.h>
#include <iostream.h>

#define __PARDEFS_H__  // prevent inclusion of "pardefs.h"
#include "frontend.h"


#define yyparse Cxx_parser::yyparse
#define YYERROR_VERBOSE

#define YYDEBUG 1    /* Cause the `yydebug' variable to be defined.  */

// more readable output for debug
#define YYPRINT(file, token, value)  cerr<< ' '<< (value.UA_token) << ')';

 
  
%}



%%

compilation_unit:
             { 
             current_definition.push(); 
             scope_for_scope_qualifier=false;
             }  
         program END_OF_COMPILATION_UNIT 
             { 
	     S.freeze($3->end_token_string);   // freeze without checking scope
	     YYACCEPT;
	     }
         ;  
 
program: /* empty */
        | extdefs
        ;

extdefs:
          lang_extdef
        | extdefs lang_extdef
                ;

extdefs_opt:
          extdefs
        | /* empty */
        ;

.hush_warning:
        ;
.warning_ok:
        ;

asm_keyword:
          ASM_KEYWORD
        | GCC_ASM_KEYWORD
        ;

lang_extdef:
          extdef
        ;

extdef:
          fndef   
        | datadef
        | template_def
        | overloaddef         { M::winternal("extdef -> overloaddef !!! (obsolete)");}
        | asm_keyword '(' string ')' ';'        { M::internal("extdef");}
        | extern_lang_string '{' extdefs_opt '}'
        | extern_lang_string .hush_warning fndef .warning_ok
        | extern_lang_string .hush_warning datadef .warning_ok
        | NAMESPACE identifier '{'
          extdefs_opt '}'              { M::internal("extdef 1"); }
        | NAMESPACE '{'
          extdefs_opt '}'              { M::internal("extdef 2"); }
        | NAMESPACE identifier '=' any_id ';'   { M::internal("extdef 3"); }
        | using_decl ';'               { M::internal("extdef 4"); }
        | USING NAMESPACE any_id ';'   { M::internal("extdef 5"); }
        ;

using_decl:
          USING qualified_id                 {reset_type_qualifier_scope();} 
        | USING global_scope qualified_id    {reset_type_qualifier_scope();}
        | USING global_scope unqualified_id  {reset_type_qualifier_scope();}
        ;

any_id:
          unqualified_id      {
                              reset_type_qualifier_scope();
                              $$.name=$1.name;
                              $$.e=$1.e;
                              $$.scope= current_scope.Get_Container();
                              }
        | qualified_id        {$$=$1; reset_type_qualifier_scope();}
        | global_scope qualified_id  {$$=$2;reset_type_qualifier_scope();}
        | global_scope unqualified_id 
                              {
                              $$.name=$2.name;
                              $$.e=$2.e;
                              $$.scope= current_scope.Get_Container();
                              reset_type_qualifier_scope();
                              }
        ;

extern_lang_string:
        EXTERN_LANG_STRING
        | extern_lang_string EXTERN_LANG_STRING  {}
        ;

template_header:
          TEMPLATE '<' 
                  {
		  Template_Entity* te= new Template_Entity(current_scope);
		  current_scope.add_Entity( te );
		  current_scope.Push( te->p );    // scope in tempalte paramters
		  $<U_Template_Entity>$=te;
		  }
          template_parm_list '>' 
                  {
		  current_scope.Pop();
		  $$=$<U_Template_Entity>3;
		  current_scope.Push( *$$ ); // scope in template
		  }
        ;

template_parm_list:
          template_parm
        | template_parm_list ',' template_parm
        ;

template_type_parm:
          aggr             { current_scope.add_Entity( new AggregateType_Entity(NULL) ); }
        | aggr identifier  { current_scope.add_Entity( new AggregateType_Entity( $2.s) ); }
        | TYPENAME_KEYWORD               {M::internal("template_type_parm 1");}
        | TYPENAME_KEYWORD identifier    {M::internal("template_type_parm 2");}
        ;

template_parm:
        /* The following rules introduce a new reduce/reduce
           conflict on the ',' and '>' input tokens: they are valid
           prefixes for a `structsp', which means they could match a
           nameless parameter.  See 14.6, paragraph 3.
           By putting them before the `parm' rule, we get
           their match before considering them nameless parameter
           declarations.  */
          template_type_parm
        | template_type_parm '=' typespec
        | full_parm
        ;

overloaddef:
          OVERLOAD ov_identifiers ';'
        ;

ov_identifiers: IDENTIFIER { }
        | ov_identifiers ',' IDENTIFIER 
        ;
          
template_def:
        /* Class template declarations go here; they aren't normal class
           declarations, because we can't process the bodies yet.  */
          template_header named_class_head_sans_basetype '{' 
         ';' {current_scope.Pop(); M::internal("template_def 1");}    /* this must not be acitevatd, see behaviour of lexer */
        | template_header named_class_head_sans_basetype_defn '{' 
             {
	     A_token t; t.tokenid=PRE_PARSED_CLASS_DECL;
	     $1->get_buffer().add_token( t );
	     $1->get_buffer().add_token( $3 ); // add the left curly barce '{'
	     
	     current_skip_buffer=&($1->get_buffer());
	     }
	  skip_till_right_brace
          ';' 
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($6->end_token_string); 
	     current_skip_buffer= &Null_skip_buffer;
	     }
        | template_header named_class_head_sans_basetype ':'
          ';' {current_scope.Pop(); M::internal("template_def 2");}    /* this must not be acitevatd,  see behaviour of lexer */
        | template_header named_class_head_sans_basetype_defn ':'
             {
	     A_token t; t.tokenid=PRE_PARSED_CLASS_DECL;
	     $1->get_buffer().add_token( t );
	     $1->get_buffer().add_token( $3 );   // add the colon
	     
	     current_skip_buffer=&($1->get_buffer());
	     }
	  skip_till_left_brace   /*skip base class*/
          skip_till_right_brace  /*skip definition of the class */
          ';' 
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($7->end_token_string);
	     current_skip_buffer= &Null_skip_buffer;
	     }
        | template_header named_class_head_sans_basetype ';'  {current_scope.Pop(); M::internal("template_def 3");}    /* this must not be acitevatd,  see behaviour of lexer */
        | template_header named_class_head_sans_basetype_defn ';'
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($3->end_token_string); 
	     }
        | template_header /* notype_initdcl0 ';' */
          notype_declarator exception_specification_opt maybeasm maybe_attribute
          fn_tmpl_end 
             { }
        | template_header typed_declspecs /*initdcl0*/
          declarator exception_specification_opt maybeasm maybe_attribute
          fn_tmpl_end 
             { }
        | template_header declmods notype_declarator fn_tmpl_end { }
        /* Try to recover from syntax errors in templates.  */
        | template_header error '}'     
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($3->end_token_string); 
	     }
        | template_header error ';'    
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($3->end_token_string); 
	     }
        ;

fn_tmpl_end: 
          '{'  
	     { current_skip_buffer= &Null_skip_buffer;} 
          skip_till_right_brace
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($3->begin_token_string); 
	     }
        | ':'           
             { 
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($1->end_token_string); 
	     }
        | ';'                
             {
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($1->end_token_string); 
	     }

        | '='           
             {
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($1->end_token_string); 
	     }

        | RETURN        
             {
	     current_scope.Pop(); 
	     current_definition.top().null(); 
	     freeze($1->end_token_string); 
	     }

        ;

datadef:
          nomods_initdecls ';'           { current_definition.top().null(); freeze($2->end_token_string); }
        | declmods notype_initdecls ';'  { current_definition.top().null(); freeze($3->end_token_string); }
        /* Normal case to make fast: "const i;".  */
        | declmods notype_declarator ';' { current_definition.top().null(); freeze($3->end_token_string); }
        | typed_declspecs  initdecls ';' { current_definition.top().null(); freeze($3->end_token_string); }
        /* Normal case: make this fast.  */
        | typed_declspecs declarator ';' { current_definition.top().null(); freeze($3->end_token_string); }
        | declmods ';'                   { current_definition.top().null(); freeze($2->end_token_string); }
        | explicit_instantiation ';'     { current_definition.top().null(); freeze($2->end_token_string); }
        | typed_declspecs ';'            { current_definition.top().null(); freeze($2->end_token_string); }
        | error ';'                      { current_definition.top().null(); freeze($2->end_token_string); }
        | error '}'                      { current_definition.top().null(); freeze($2->end_token_string); }
        | ';'                            { current_definition.top().null(); freeze($1->end_token_string); }
        ;

ctor_initializer_opt:
          nodecls
        | base_init
        ;

maybe_return_init:
          /* empty */
        | return_init
        | return_init ';'
        ;

eat_saved_input:
          /* empty */
        | END_OF_SAVED_INPUT
        ;



fndef:
          fn.def1 maybe_return_init ctor_initializer_opt 
                {
		Grouping_Entity * g;
		switch( $1->cl )
		    {
		    case Entity::Function_E : 
		         g= &( (Function_Entity*)$1 )  ->        Get_local(); break;
		    case Entity::Particular_overloadable_E :
		         g= &( (Particular_overloadable*)$1 ) -> Get_local(); break;
		    default:
		         M::winternal("fndef: no function" );
		         YYERROR;
		         break;
		    }
		current_scope.Push(*g);
		} 
          compstmt_or_error
                { current_scope.Pop(); }
        | fn.def1 maybe_return_init function_try_block  eat_saved_input { M::internal("fndef: eat_saved_input ???");}
        | typed_declspecs declarator error {}
        | declmods notype_declarator error {}
        | notype_declarator error {}
        ;

fn.def1:
          typed_declspecs declarator exception_specification_opt {$$=$2;}
        | declmods notype_declarator exception_specification_opt { declare($2); $$=$2.e;  M::winternal("fn.def1 1");}
        | notype_declarator exception_specification_opt { declare($1); $$=$1.e; M::winternal("fn.def1 2");}
        | PRE_PARSED_FUNCTION_DECL { M::internal("fn.def1 3");}
        ;

/* more C++ complexity.  See component_decl for a comment on the
   reduce/reduce conflict introduced by these rules.  */
fn.def2:
                       /* here we have function with type but without name (because there isn't declarator) */
          typed_declspecs begin_parmlist parmlist ')' type_quals exception_specification_opt
              {
              end_parameters();
              Function *f= $2;
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
	      f->set_inner( current_definition.top().type );   // constructor for a type is a  function returning the same type
                
              A_name_entity_type i;
              i.name= new Entity_Name();
              i.t=f;
              i.e= NULL;                    // function with no name is recognized by declare(...)
              declare(i);
              delete i.name;
	      $$=i.e;
              }
        | typed_declspecs LEFT_RIGHT type_quals exception_specification_opt     
              {
              Function *f= new Function( current_scope.Get_Container() );
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              f->set_inner( current_definition.top().type );   // constructor for a type is a  function returning the same type
                
              A_name_entity_type i;
              i.name= new Entity_Name();
              i.t=f;
              i.e= NULL;                    // function with no name is recognized by declare(...)
              declare(i);
              delete i.name;
	      $$=i.e;
              }

        | typed_declspecs declarator exception_specification_opt { $$= $2; }
        | declmods notype_declarator exception_specification_opt 
              {
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              $2.t->set_inner( current_definition.top().type);
              declare($2);
              delete $2.name;
	      $$= $2.e;
	      M::winternal("fn.def2 3");
              }
        | notype_declarator exception_specification_opt 
              {
              $1.t->set_inner(new Built_in(Void) );
              declare($1);
              delete $1.name;
	      $$=$1.e;
              }
        ;

return_id: RETURN IDENTIFIER {}
        ;

return_init: return_id maybe_init
        | return_id '(' nonnull_exprlist ')' { delete_ListTypes( $3 ); }
        | return_id LEFT_RIGHT
        ;

base_init:
          ':' .set_base_init member_init_list {}
        ;

.set_base_init:
        /* empty */
        ;

member_init_list:
          /* empty */
        | member_init
        | member_init_list ',' member_init
        | member_init_list error
        ;

member_init: '(' nonnull_exprlist ')' { delete_ListTypes( $2 ); }
        | LEFT_RIGHT                                  { }
        | notype_identifier '(' nonnull_exprlist ')'  { delete_ListTypes( $3 ); }
        | notype_identifier LEFT_RIGHT                { }
        | complete_type_name '(' nonnull_exprlist ')' {  delete_ListTypes( $3 );}
        | complete_type_name LEFT_RIGHT  {}
        /* GNU extension */
        | notype_qualified_id '(' nonnull_exprlist ')' {  delete_ListTypes( $3 ); M::winternal("member_init 1");}
        | notype_qualified_id LEFT_RIGHT  {M::winternal("member_init 2");}
        ;

identifier:
          IDENTIFIER {
                     reset_type_qualifier_scope();
                     $$.s=$1->aux.id.s; $$.e=$1->aux.id.e;
                     }
        | TYPENAME   { 
                     reset_type_qualifier_scope();
                     $$.s=$1->aux.id.s; $$.e=$1->aux.id.e; 
                     }
        | PTYPENAME  { 
                     reset_type_qualifier_scope();
                     $$.s=$1->aux.id.s; $$.e=$1->aux.id.e;
                     }
        | NSNAME     
                     { 
                     reset_type_qualifier_scope();
                     $$.s=$1->aux.id.s; $$.e=$1->aux.id.e; 
                     }
        ;

notype_identifier:
          IDENTIFIER     { $$.s=$1->aux.id.s; $$.e=$1->aux.id.e; }
        | PTYPENAME      { $$.s=$1->aux.id.s; $$.e=$1->aux.id.e;  }
        | NSNAME %prec EMPTY { $$.s=$1->aux.id.s; $$.e=$1->aux.id.e; }
        ;

identifier_defn:
          IDENTIFIER_DEFN { $$.s=$1->aux.id.s; $$.e=$1->aux.id.e; }
        | TYPENAME_DEFN   { $$.s=$1->aux.id.s; $$.e=$1->aux.id.e; }
        | PTYPENAME_DEFN  { $$.s=$1->aux.id.s; $$.e=$1->aux.id.e;   }
        ;

explicit_instantiation:
          TEMPLATE specialization  template_instantiation { M::internal("explicit_instantiation 1"); }
        | TEMPLATE typed_declspecs declarator
        | TEMPLATE notype_declarator
        | SCSPEC TEMPLATE specialization template_instantiation  { M::internal("explicit_instantiation 2"); }
        | SCSPEC TEMPLATE typed_declspecs declarator {}
        | SCSPEC TEMPLATE notype_declarator {}
        ;

template_type:
          template_type_name tmpl.2
                  { 
		  Entity* e= current_scope.Name_LookUp( *$1.n );
		  template_begin_instantiation( e, *$1.n );
		  current_scope.Push(*(AggregateType_Entity*)e);
		  current_definition.push();
		  delete $1.n;
		  $<U_Entity>$=e;
		  }
          template_instantiation 
                  {
		  $$.e= (Template_Entity*)$<U_Entity>3;
		  $$.end_string= $1.end_string;
		  current_definition.pull();
		  current_scope.Pop();
//		  ostrstream s;
//		  s << "end instatiation: "; $$.e->simple_output(s, 0 ); s << ends;
//		  M::warning(99, s.str() );
//		  delete s.str();
		  }
        ;

template_type_name:
          PTYPENAME '<' template_arg_list '>' 
             {
	     $$.n= new Entity_Name( (Template_Entity*)$1->aux.id.e, $3);
	     $$.end_string= $4->end_token_string;
	     }
        | PTYPENAME '<' '>'                   {M::internal("template_type_name 1");  /* what a strage syntax */ }
        | TYPENAME  '<' template_arg_list '>' {M::internal("template_type_name 2");}
        ;

tmpl.2: 
          /* empty */ %prec EMPTY
        ;

template_arg_list:
          template_arg                       { $$= new list<Type*>; $$->push_back( $1 ); }
        | template_arg_list ',' template_arg { $$= $1; $$->push_back( $3 ); }
        ;

template_arg:
          type_id                     { $$= $1; }
        | expr_no_commas  %prec UNARY 
		 {  
		 // non type parameters are not supported yet.   %type<..> must be modified
		 // But it can be also a struct name (WARNING: I don't understand this!) 
		 $$= new Built_in(Void);    // a fictitious type to avoid error
		 M::winternal("template_arg 2");
		 }
        ;

template_instantiate_once:
          PRE_PARSED_CLASS_DECL maybe_base_class_list
          left_curly opt.component_decl_list '}' { pop_source(); }
	 | PRE_PARSED_CLASS_DECL error END_PRE_PARSED { pop_source(); }
        ;

template_instantiation:
          /* empty */
        | template_instantiate_once
        ;

template_instantiate_some:
          /* empty */
	   /* the followin roule is for instatiate function templates */
/*        | template_instantiate_once template_instantiate_some */
        ;

unop:     '-'      { $$=$1; }
        | '+'      { $$=$1; }
        | PLUSPLUS { $$=$1; }
        | MINUSMINUS { $$=$1; }
        | '!'      { $$=$1; }
        ;

expr:     nontrivial_exprlist 
	     { 
	     $$.t= *(--$1->end()); 
	     $1->erase(--$1->end());  
	     delete_ListTypes( $1 ); 
	     }
        | expr_no_commas
        ;

paren_expr_or_null:
        LEFT_RIGHT     {}
        | '(' expr ')' {}
        ;

paren_cond_or_null:
        LEFT_RIGHT           {}
        | '(' condition ')'  {}
        ;

xcond:
        /* empty */
        | condition
        | error
        ;

condition:
        type_specifier_seq declarator exception_specification_opt maybeasm maybe_attribute '='
        init {M::internal("condition");}
        | expr {}
        ;

compstmtend:
          '}'                            { current_definition.top().null(); freeze($1->end_token_string); }
        | maybe_label_decls stmts '}'    { current_definition.top().null(); freeze($3->end_token_string); }
        | maybe_label_decls stmts error '}' { current_definition.top().null(); freeze($4->end_token_string); }
        | maybe_label_decls error '}'    { current_definition.top().null(); freeze($3->end_token_string); }
        ;

already_scoped_stmt:
          '{' compstmtend {}
        | simple_stmt
        ;


nontrivial_exprlist:
          expr_no_commas ',' expr_no_commas      
	       { 
	       $$= new list<Type*>;
	       $$->push_back($1.t);
	       $$->push_back($3.t);
	       }
        | expr_no_commas ',' error
	       { 
	       $$= new list<Type*>;
	       $$->push_back($1.t);
	       }
        | nontrivial_exprlist ',' expr_no_commas
	       { 
	       $$= $1;
	       $$->push_back($3.t);
	       }
        | nontrivial_exprlist ',' error {$$=$1}
        ;

nonnull_exprlist:
          expr_no_commas 
	     {
	     $$= new list<Type*>;
	     $$->push_back($1.t);
	     }
        | nontrivial_exprlist {$$=$1}
        ;

unary_expr:
          primary %prec UNARY
        /* __extension__ turns off -pedantic for following primary.  */
        | EXTENSION  cast_expr %prec UNARY { $$.t= $2.t; $$.begin=$1.begin_token_string; $$.end=$2.end; }
        | '*' cast_expr   %prec UNARY { $$.t= applay_operator($2.t, $1);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        | '&' cast_expr   %prec UNARY { $$.t= applay_operator($2.t, $1);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        | '~' cast_expr               { $$.t= applay_operator($2.t, $1);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        | unop cast_expr  %prec UNARY { $$.t= applay_operator($2.t, $1);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        /* Refer to the address of a label as a pointer.  */
        | ANDAND identifier    { M::winternal("unary_expr 1"); $$.t= new Built_in(Error);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        | SIZEOF unary_expr  %prec UNARY            { $$.t= new Built_in(Int);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        | SIZEOF '(' type_id ')'  %prec HYPERUNARY  { $$.t= new Built_in(Int);  $$.begin=$1.begin_token_string; $$.end=$4.end_token_string;}
        | ALIGNOF unary_expr  %prec UNARY             { M::winternal("unary_expr: alingof() 1");  $$.t= new Built_in(Int);  $$.begin=$1.begin_token_string; $$.end=$2.end;}
        | ALIGNOF '(' type_id ')'  %prec HYPERUNARY   { M::winternal("unary_expr: alingof() 2");  $$.t= new Built_in(Int);  $$.begin=$1.begin_token_string; $$.end=$4.end_token_string;}

        /* The %prec EMPTY's here are required by the = init initializer
           syntax extension; see below.  */
        | new new_type_id %prec EMPTY                    { $$.t= new Pointer(); $$.t->set_inner( $2.t->clone() );  $$.begin=$1.begin; $$.end=$2.end;}
        | new new_type_id new_initializer                { $$.t= new Pointer(); $$.t->set_inner( $2.t->clone() );  $$.begin=$1.begin; $$.end=$3.end;}
        | new new_placement new_type_id %prec EMPTY      { $$.t= new Pointer(); $$.t->set_inner( $3.t->clone() );  $$.begin=$1.begin; $$.end=$3.end;}
        | new new_placement new_type_id new_initializer  { $$.t= new Pointer(); $$.t->set_inner( $3.t->clone() );  $$.begin=$1.begin; $$.end=$4.end;}
        | new '(' type_id ')' %prec EMPTY                { $$.t= new Pointer(); $$.t->set_inner( $3->clone() );  $$.begin=$1.begin; $$.end=$4.end_token_string;}
        | new '(' type_id ')' new_initializer            { $$.t= new Pointer(); $$.t->set_inner( $3->clone() );  $$.begin=$1.begin; $$.end=$5.end;}
        | new new_placement '(' type_id ')' %prec EMPTY  { $$.t= new Pointer(); $$.t->set_inner( $4->clone() );  $$.begin=$1.begin; $$.end=$5.end_token_string;}
        | new new_placement '(' type_id ')' new_initializer {$$.t= new Pointer();$$.t->set_inner( $4->clone() ); $$.begin=$1.begin; $$.end=$6.end;}
        | delete cast_expr  %prec UNARY                  { $$.t= new Built_in(Void);  $$.begin=$1.begin; $$.end=$2.end;}
        | delete '[' ']' cast_expr  %prec UNARY      { $$.t= new Built_in(Void);  $$.begin=$1.begin; $$.end=$4.end;}
        | delete '[' expr ']' cast_expr %prec UNARY  { $$.t= new Built_in(Void); delete $3.t;  $$.begin=$1.begin; $$.end=$5.end;}
        ;

new_placement:
          '(' nonnull_exprlist ')' { delete_ListTypes( $2 );}
        | '{' nonnull_exprlist '}' { delete_ListTypes( $2 );}
        ;

new_initializer:
          '(' nonnull_exprlist ')' {  delete_ListTypes( $2 ); }
        | LEFT_RIGHT               {}
        | '(' typespec ')'         { $$.begin= $1.begin_token_string; $$.end= $3.end_token_string; }
        /* GNU extension so people can use initializer lists.  Note that
           this alters the meaning of `new int = 1', which was previously
           syntactically valid but semantically invalid.  */
        | '=' init  { $$.begin= $1.begin_token_string; $$.end= $2.end; }
        ;

/* This is necessary to postpone reduction of `int ((int)(int)(int))'.  */
regcast_or_absdcl:
          '(' type_id ')' %prec EMPTY     { $$.t=$2; $$.begin= $1.begin_token_string; $$.end= $3.end_token_string;}
        | regcast_or_absdcl '(' type_id ')' %prec EMPTY  {M::internal("regcast_or_absdcl"); $$.begin= $1.begin; $$.end= $4.end_token_string;}
        ;

cast_expr:
          unary_expr            { $$.t= $1.t; $$.begin= $1.begin; $$.end= $1.end; }
        | regcast_or_absdcl unary_expr  %prec UNARY   {$$.t= $1.t; $$.begin= $1.begin; $$.end= $2.end;}
        | regcast_or_absdcl '{' initlist maybecomma '}'  %prec UNARY {$$.t= $1.t; $$.begin= $1.begin; $$.end= $5.end_token_string;}
        ;

expr_no_commas:
          cast_expr 
        /* Handle general members.  */
        | expr_no_commas POINTSAT_STAR expr_no_commas
        | expr_no_commas DOT_STAR expr_no_commas
        | expr_no_commas '+' expr_no_commas 
        | expr_no_commas '-' expr_no_commas
        | expr_no_commas '*' expr_no_commas
        | expr_no_commas '/' expr_no_commas
        | expr_no_commas '%' expr_no_commas
        | expr_no_commas LSHIFT expr_no_commas
        | expr_no_commas RSHIFT expr_no_commas
        | expr_no_commas ARITHCOMPARE expr_no_commas
        | expr_no_commas '<' expr_no_commas
        | expr_no_commas '>' expr_no_commas
        | expr_no_commas EQCOMPARE expr_no_commas
        | expr_no_commas MIN_MAX expr_no_commas
        | expr_no_commas '&' expr_no_commas
        | expr_no_commas '|' expr_no_commas
        | expr_no_commas '^' expr_no_commas
        | expr_no_commas ANDAND expr_no_commas
        | expr_no_commas OROR expr_no_commas
        | expr_no_commas '?' xexpr ':' expr_no_commas
        | expr_no_commas '=' expr_no_commas
        | expr_no_commas ASSIGN expr_no_commas
        | THROW                 { $$.t= new Built_in(Void); }
        | THROW expr_no_commas  { $$.t= new Built_in(Void); }
/* These extensions are not defined.  The second arg to build_m_component_ref
   is old, build_m_component_ref now does an implicit
   build_indirect_ref (x, NULL_PTR) on the second argument.*/
      /*
        | object '&' expr_no_commas   %prec UNARY { M::internal("expr_no_commas: 1");}
        | object unop expr_no_commas  %prec UNARY { M::internal("expr_no_commas: 2"); }
        | object '(' type_id ')' expr_no_commas  %prec UNARY { M::internal("expr_no_commas: 3");}
        | object primary_no_id  %prec UNARY                  { M::internal("expr_no_commas: 4");  }
       */
        ;

notype_unqualified_id:
          '~' see_typename identifier 
                  { 
                  $$.name= new Entity_Name( Entity_Name::destructor, $3.s );
                  $$.e= current_scope.Get_Container()->Local_Name_LookUp( *$$.name);
		  $$.begin= $1.begin_token_string; $$.end= $3.end; 
                  }
        | operator_name
                  {
                  $$.name= new Entity_Name(*$1.oi); delete $1.oi;
                  $$.e=current_scope.Name_LookUp(*$$.name);
		  $$.begin= $1.begin; $$.end= $1.end;
                  }    
        | IDENTIFIER   
                  {
                  $$.name= new Entity_Name( $1->aux.id.s );
                  $$.e= $1->aux.id.e;
		  $$.begin= $1.begin_token_string; $$.end= $1.end_token_string;
                  }
        | PTYPENAME         
                  {
                  $$.name= new Entity_Name( $1->aux.id.s );
                  $$.e= $1->aux.id.e;
		  $$.begin= $1.begin_token_string; $$.end= $1.end_token_string;
                  }
        | NSNAME %prec EMPTY 
                  { 
		  M::internal("notype_unqualified_id 4");
		  $$.begin= $1.begin_token_string; $$.end= $1.end_token_string;
		  }
        ;

unqualified_id:
          notype_unqualified_id { $$=$1; }
        | TYPENAME               
                  {
                  $$.name= new Entity_Name( $1->aux.id.s );
                  $$.e= $1->aux.id.e;
		  $$.begin= $1.begin_token_string; $$.end= $1.end_token_string;
                  }
        ;

expr_or_declarator:
          notype_unqualified_id 
                   {
                   $$.name=$1.name;
                   $$.e=$1.e;
                   $$.t=new Direct;
                   }
        | '*' expr_or_declarator %prec UNARY {$$= $2; $$.t->set_inner( new Pointer);}
        | '&' expr_or_declarator %prec UNARY {$$= $2; $$.t->set_inner( new Reference);}
        | '(' expr_or_declarator ')'         {$$= $2; }
        ;

direct_notype_declarator:
          complex_direct_notype_declarator {$$=$1;}
        | notype_unqualified_id 
                   {
                   $$.name=$1.name;
                   $$.e=$1.e;
                   $$.t=new Direct;
		   $$.scope= current_scope.Get_Container();
                   }
        | '(' expr_or_declarator ')' 
                   {
                   $$.name=$2.name;
                   $$.e=$2.e;
                   $$.t=$2.t;
		   $$.scope= current_scope.Get_Container();
                   }
        ;

primary:
          notype_unqualified_id { $$.t= get_type_from_entity( $1.e ); $$.begin= $1.begin; $$.end= $1.end;}
        | CONSTANT           { $$.t= new Built_in(Int); $$.begin= $1.begin_token_string;  $$.end= $1.end_token_string  }
        | boolean.literal    { $$.t= new Built_in(Int); $$.begin=$1.begin; $$.end=$1.end; }
        | string             { $$.t= new Pointer(); $$.t->set_inner(new Built_in(Char)); $$.begin=$1.begin; $$.end=$1.end;}
        | '(' expr ')'       { $$.t= $2.t; $$.begin=$1.begin_token_string; $$.end=$3.end_token_string; }
        | '(' expr_or_declarator ')' 
	                     {
			     if (!$2.e)
			        {
			        M::hl_error(M::Undeclared_symbol);
				$$.t= new Built_in(Error);
				}
			     else
			        $$.t= convert_expr_or_declarator($2); 
			     
			     delete $2.t;
			     delete $2.name;
			     
			     $$.begin=$1.begin_token_string; $$.end=$3.end_token_string;       
			     }
        | '(' error ')'      { $$.t= new Built_in(Error);  $$.begin=$1.begin_token_string; $$.end=$3.end_token_string;  }
        | '(' compstmt ')'      { M::warning( 4, "primary: g++ extension, {..} inside expression "); }    /* g++ extension */
        | primary '(' nonnull_exprlist ')' template_instantiate_some
	                     {
	                     $$.t= apply_function( $1.t, $3 );
	                     delete_ListTypes($3);
			     $$.begin=$1.begin;
			     $$.end=$4.end_token_string; 
	                     }                
        | primary LEFT_RIGHT    
	                     {
	                     list<Type*>* L= new list<Type*>;
	                     $$.t= apply_function( $1.t, L );
	                     delete_ListTypes(L);
			     $$.begin=$1.begin; 
			     $$.end=$2.end_token_string;
	                     }
        | primary '[' expr ']'  { $$.t= applay_operator($1.t, $2 ); delete $3.t; $$.begin=$1.begin; $$.end= $4.end_token_string; }
        | primary PLUSPLUS     { $$= $1; $$.begin=$1.begin; $$.end= $2.end_token_string;}
        | primary MINUSMINUS   { $$= $1; $$.begin=$1.begin; $$.end= $2.end_token_string;}
        /* C++ extensions */
        | THIS                 
	       { 
	       AggregateType_Entity* a= get_current_class();
	       if (a )
		   {
	           $$.t= new Pointer;
	           $$.t->set_inner( new Aggregate(a )  );
		   }
	       else
		   $$.t= new Built_in(Error);
	       $$.begin= $1.begin_token_string;$$.end= $1.end_token_string;
	       }
        | TYPE_QUAL '(' nonnull_exprlist ')' { $$.t= new Built_in(Int); $$.begin= $1.begin_token_string;$$.end= $4.end_token_string; }
        | functional_cast       
        | DYNAMIC_CAST '<' type_id '>' '(' expr ')'     { $$.t= $3; delete $6.t; $$.begin= $1.begin_token_string;$$.end= $7.end_token_string; }
        | STATIC_CAST '<'  type_id '>' '(' expr ')'     { $$.t= $3; delete $6.t; $$.begin= $1.begin_token_string;$$.end= $7.end_token_string;}
        | REINTERPRET_CAST '<' type_id '>' '(' expr ')' { $$.t= $3; delete $6.t; $$.begin= $1.begin_token_string;$$.end= $7.end_token_string;}
        | CONST_CAST '<'  type_id '>' '(' expr ')'      { $$.t= $3; delete $6.t; $$.begin= $1.begin_token_string;$$.end= $7.end_token_string;}
        | TYPEID '(' expr ')'       { M::winternal("primary: typeid() 1"); $$.t= new Built_in(Error); delete $3.t; $$.begin= $1.begin_token_string;$$.end= $4.end_token_string;}
        | TYPEID '(' type_id ')'    { M::winternal("primary: typeid() 2"); $$.t= new Built_in(Error); delete $3; $$.begin= $1.begin_token_string;$$.end= $4.end_token_string;}
        | global_scope IDENTIFIER    { reset_type_qualifier_scope(); $$.t= get_type_from_entity( $2.aux.id.e ); $$.begin= $1.begin; $$.end= $2.end_token_string;}
        | global_scope operator_name 
	         { 
		 reset_type_qualifier_scope();
		 current_scope.Push(*Global);
                 Entity* e=current_scope.Name_LookUp(Entity_Name(*$2.oi));
		 current_scope.Pop();
		 $$.t=get_type_from_entity(e);
		 delete $2.oi;
		 $$.begin= $1.begin; $$.end= $2.end;
		 }
        | overqualified_id %prec HYPERUNARY         { $$.t= get_type_from_entity($1.e); $$.begin= $1.begin; $$.end= $1.end; }
        | overqualified_id '(' nonnull_exprlist ')' 
	    {
	    Type *t= get_type_from_entity($1.e);
	    $$.t= apply_function( t, $3 );
	    delete_ListTypes($3);
	    $$.begin= $1.begin; $$.end= $4.end_token_string;
	    }
        | overqualified_id LEFT_RIGHT               
	         {
	         Type *t= get_type_from_entity($1.e);
	         list<Type*>* L= new list<Type*>;
	         $$.t= apply_function( t, L );
	         delete_ListTypes(L);
	         $$.begin= $1.begin; $$.end= $2.end_token_string;
	         }
	| primary POINT_LEFTBRACE complete_type_name '}'
	         {
		 A_object ob;
	         ob.k= A_object::get_pointer;
	         ob.primary=$1;
	         ob.e= $3;
	         ob.begin_oper=$2.begin_token_string; 
	         ob.end_oper=$4.end_token_string;
		 A_name_entity fictitious;
		 $$.t= modify_access(ob,  fictitious);
	         $$.begin= $1.begin; $$.end=$4.end_token_string;
	         }   
        | object unqualified_id  %prec UNARY             
	         { $$.t= modify_access($1, $2 ); $$.begin= $1.begin; $$.end= $2.end; }
        | object overqualified_id %prec UNARY            
	         { $$.t= modify_access($1, $2); $$.begin= $1.begin; $$.end= $2.end;}
        | object unqualified_id '(' nonnull_exprlist ')' 
	         { 
	         Type* t= modify_access($1, $2 );
	         $$.t= apply_function( $1.primary.t, $4 );
		 delete_ListTypes($4);
	         $$.begin= $1.begin; $$.end= $5.end_token_string;
	         }    
        | object unqualified_id LEFT_RIGHT 
	         {  
	         Type* t= modify_access($1, $2);
		 list<Type*>* L= new list<Type*>;
	         $$.t= apply_function( t, L );
		 delete_ListTypes(L);
	         $$.begin= $1.begin; $$.end= $2.end;
	         }    
        | object overqualified_id '(' nonnull_exprlist ')' 
	         {  
	         Type* t= modify_access($1, $2);
	         $$.t= apply_function( t, $4 );
		 delete_ListTypes($4);
	         $$.begin= $1.begin; $$.end= $5.end_token_string;
	         }
        | object overqualified_id LEFT_RIGHT  
	         {  
	         Type* t= modify_access($1, $2);
		 list<Type*>* L= new list<Type*>;
	         $$.t= apply_function( t, L );
		 delete_ListTypes(L);
	         $$.begin= $1.begin; $$.end= $3.end_token_string;
	         }
        | object '~' TYPESPEC LEFT_RIGHT     /* p->int::~int() is valid -- 12.4 */
	         {$$.t= new Built_in(Void); $$.begin= $1.begin; $$.end= $4.end_token_string; }
        | object TYPESPEC SCOPE '~' TYPESPEC LEFT_RIGHT 
	         {$$.t= new Built_in(Void); $$.begin= $1.begin; $$.end= $6.end_token_string;}
        | object error   
	         {$$.t= new Built_in(Error); $$.begin= $1.begin; }
        ;

/*
* primary_no_id:
*           '(' expr ')'     { $$= $2 }
*         | '(' error ')'    { $$.t= new Built_in(Error); }
*         | '(' compstmt ')'    
*             {  // g++ extension
* 	    $$.t= new Built_in(Error); 
* 	    M::winternal("primary_no_id g++ extension");
* 	    }
*         | primary_no_id '(' nonnull_exprlist ')'  { $$.t= new Built_in(Error); }
*         | primary_no_id LEFT_RIGHT                { $$.t= new Built_in(Error); }
*         | primary_no_id '[' expr ']'              { $$.t= new Built_in(Error); delete $3.t;}
*         | primary_no_id PLUSPLUS           { $$= $1 }
*         | primary_no_id MINUSMINUS          { $$= $1 }
*         | SCOPE IDENTIFIER               { M::winternal("primary_no_id 1"); $$.t= new Built_in(Error); }
*         | SCOPE operator_name            { M::winternal("primary_no_id 2"); $$.t= new Built_in(Error); }
*         ;
*/

new:      NEW              { $$.begin= $1.begin_token_string; $$.end= $1.end_token_string; }
        | global_scope NEW {reset_type_qualifier_scope(); $$.begin= $1.begin; $$.end= $2.end_token_string;}
        ;

delete:   DELETE           { $$.begin= $1.begin_token_string; $$.end= $1.end_token_string; }
        | global_scope DELETE {reset_type_qualifier_scope(); $$.begin= $1.begin; $$.end= $2.end_token_string;}
        ;

boolean.literal:
          CXX_TRUE { $$.begin= $1.begin_token_string; $$.end= $1.end_token_string; }
        | CXX_FALSE { $$.begin= $1.begin_token_string; $$.end= $1.end_token_string; }
        ;

/* Produces a STRING_CST with perhaps more STRING_CSTs chained onto it.  */
string:
          STRING         {$$.begin= $1.begin_token_string; $$.end= $1.end_token_string; }
        | string STRING  {$$.begin= $1.begin; $$.end= $2.end_token_string; }
        ;

nodecls:
          /* empty */
        ;

object:   primary '.'                
	     {
	     $$.k= A_object::point; 
	     $$.primary=$1; 
	     $$.begin_oper=$2.begin_token_string; 
	     $$.end_oper=$2.end_token_string;
	     $$.begin= $1.begin; $$.end=$2.end_token_string;
	     }
        | primary POINTSAT           
	     {
	     $$.k= A_object::pointsat; 
	     $$.primary=$1; 
	     $$.begin_oper=$2.begin_token_string; 
	     $$.end_oper=$2.end_token_string;
	     $$.begin= $1.begin; $$.end=$2.end_token_string;
	     }
	  
	 /* ECO extensions */
        | primary POINT_LEFTBRACE complete_type_name RIGHTBRACE_POINT
	     {
	     $$.k= A_object::point_dynamic; 
	     $$.primary=$1; 
	     $$.e= $3;
	     $$.begin_oper=$2.begin_token_string; 
	     $$.end_oper=$4.end_token_string;
	     $$.begin= $1.begin; $$.end=$4.end_token_string;
	     }
        | primary   MINUS_LEFTBRACE complete_type_name RIGHTBRACE_GREATER
	     {
	     $$.k= A_object::pointsat_dynamic;
	     $$.primary=$1; 
	     $$.e= $3;
	     $$.begin_oper=$2.begin_token_string; 
	     $$.end_oper=$4.end_token_string;
	     $$.begin= $1.begin; $$.end=$4.end_token_string;
	     }
        ;

decl:
        /* Normal case: make this fast.  */
          typespec declarator ';'  { current_definition.top().null();}
        | typed_declspecs declarator ';'  { current_definition.top().null();}
        | typespec initdecls ';'  { current_definition.top().null();}
        | typed_declspecs initdecls ';'  { current_definition.top().null();}
        | declmods notype_initdecls ';'  { current_definition.top().null();}
        | typed_declspecs ';'  { current_definition.top().null();}
        | declmods ';'  { current_definition.top().null();}
        ;

/* Any kind of declarator (thus, all declarators allowed
   after an explicit typespec).  */

declarator:
          after_type_declarator %prec EMPTY 
              {
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              $1.t->set_inner( current_definition.top().type);
              declare($1);
              delete $1.name;
	      $$= $1.e;
              }
        | notype_declarator %prec EMPTY
              {
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              $1.t->set_inner( current_definition.top().type);
              declare($1);
              delete $1.name;
	      $$= $1.e;
              }
        ;

/* This is necessary to postpone reduction of `int()()()()'.  */
fcast_or_absdcl:
          LEFT_RIGHT %prec EMPTY                  { $$.begin= $1.begin_token_string; $$.end= $1.end_token_string;}
        | fcast_or_absdcl LEFT_RIGHT %prec EMPTY  { $$.begin= $1.begin;              $$.end= $2.end_token_string;}
        ;

/* ANSI type-id (8.1) */
type_id:
          typed_typespecs absdcl      
	      {
	      $$=$2; 
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
	      $$->set_inner(current_definition.top().type); 
	      }
        | nonempty_type_quals absdcl  {$$=$2; $$->set_inner(new Built_in(Int) );}
        | typespec absdcl             
	       {
	       $$=$2; 
 	       if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
	       $$->set_inner(current_definition.top().type); 
	       }
        | typed_typespecs %prec EMPTY 
	       {
	       if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
	       $$=current_definition.top().type; 
	       }
        | nonempty_type_quals %prec EMPTY  {$$= new Built_in(Int);}
        ;

/* Declspecs which contain at least one type specifier or typedef name.
   (Just `const' or `volatile' is not enough.)
   A typedef'd name following these is taken as a name to be declared.
   In the result, declspecs have a non-NULL TREE_VALUE, attributes do not.  */

typed_declspecs:
          typed_typespecs %prec EMPTY 
        | typed_declspecs1    
        ;

typed_declspecs1:
          declmods typespec   
        | typespec reserved_declspecs   %prec HYPERUNARY  
        | typespec reserved_typespecquals reserved_declspecs 
        | declmods typespec reserved_declspecs 
        | declmods typespec reserved_typespecquals 
        | declmods typespec reserved_typespecquals reserved_declspecs 
        ;

reserved_declspecs:
          SCSPEC                         { (int)current_definition.top().sc|= $1->aux.sc;  }
        | reserved_declspecs typespecqual_reserved
        | reserved_declspecs SCSPEC      { (int)current_definition.top().sc|= $2->aux.sc;  }
        | reserved_declspecs attributes  { M::warning( 4, "g++ extension attribute used" ); }
        | attributes                     { M::warning( 4, "g++ extension attribute used" ); }
        ;

/* List of just storage classes and type modifiers.
   A declaration can start with just this, but then it cannot be used
   to redeclare a typedef-name.
   In the result, declspecs have a non-NULL TREE_VALUE, attributes do not.  */

declmods:
          nonempty_type_quals %prec EMPTY { (int)current_definition.top().tq|= $1; }
        | SCSPEC               { (int)current_definition.top().sc|= $1->aux.sc; }
        | declmods TYPE_QUAL   { (int)current_definition.top().tq|= $2->aux.tq; }
        | declmods SCSPEC      { (int)current_definition.top().sc|= $2->aux.sc;  }
        | declmods attributes  { M::warning( 4, "g++ extension attribute used" ); }
        | attributes           { M::warning( 4, "g++ extension attribute used" ); }	   
        ;

/* Used instead of declspecs where storage classes are not allowed
   (that is, for typenames and structure components).

   C++ can takes storage classes for structure components.
   Don't accept a typedef-name if anything but a modifier precedes it.  */

typed_typespecs:
          typespec  %prec EMPTY  
        | nonempty_type_quals typespec {(int)current_definition.top().tq|=$1;}
        | typespec reserved_typespecquals 
        | nonempty_type_quals typespec reserved_typespecquals {(int)current_definition.top().tq|=$1;}
        ;

reserved_typespecquals:
          typespecqual_reserved 
        | reserved_typespecquals typespecqual_reserved
        ;

/* A typespec (but not a type qualifier).
   Once we have seen one of these in a declaration,
   if a typedef name appears then it is being redeclared.  */

typespec: structsp               { current_definition.top().type= new Aggregate($1); }
        | TYPESPEC  %prec EMPTY  { current_definition.top().type= new Built_in($1->aux.ts); }
        | complete_type_name     {
                                 Type* t;
                                 switch( $1.e->cl )
                                      {
                                      case Entity::Simple:
                                          t= ((Simple_Entity*)$1.e)->Get_Type()->clone();
                                          break;
				      case Entity::Overloadable:
				          t= (*((Overloadable_Entity*)$1.e)->Get_Functions().begin())->T->clone();
				          break;
                                      case Entity::Enumeration:
                                      case Entity::AggregateType:
                                          t= new Aggregate( (AggregateType_Entity *)$1.e );
                                          break;
                                      default:
                                          t= NULL;
                                          M::internal("typespec: TYPENAME refer to something that can't be a type");
                                          break;
                                      }
                                 current_definition.top().type= t;
                                 }
        | TYPEOF '(' expr ')'    { M::internal("typespec 1"); }
        | TYPEOF '(' type_id ')' { M::internal("typespec 2"); }
        | SIGOF '(' expr ')'     { M::internal("typespec 3"); }
        | SIGOF '(' type_id ')'  { M::internal("typespec 4"); }
        ;

/* A typespec that is a reserved word, or a type qualifier.  */

typespecqual_reserved:
          TYPESPEC   { current_definition.top().type= new Built_in($1->aux.ts); }
        | TYPE_QUAL  { current_definition.top().tq= $1->aux.tq; }
        | structsp   { current_definition.top().type= new Aggregate($1); }
        ;

initdecls:
          initdcl0
        | initdecls ',' initdcl
        ;

notype_initdecls:
          notype_initdcl0
        | notype_initdecls ',' initdcl
        ;

nomods_initdecls:
          nomods_initdcl0
        | nomods_initdecls ',' initdcl
        ;

maybeasm:
          /* empty */
        | asm_keyword '(' string ')'
        ;

initdcl0:
          declarator exception_specification_opt maybeasm maybe_attribute '='
          init {}
/* Note how the declaration of the variable is in effect while its init is parsed! */
        | declarator exception_specification_opt maybeasm maybe_attribute {}
        ;

initdcl:
          declarator exception_specification_opt maybeasm maybe_attribute '='
          init {}
/* Note how the declaration of the variable is in effect while its init is parsed! */
        | declarator exception_specification_opt maybeasm maybe_attribute {}
        ;

notype_initdcl0:
          notype_declarator exception_specification_opt maybeasm maybe_attribute '='
          init 
	      {
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              $1.t->set_inner( current_definition.top().type);
              declare($1);
              delete $1.name;
	      }
/* Note how the declaration of the variable is in effect while its init is parsed! */
        | notype_declarator exception_specification_opt maybeasm maybe_attribute 
	      {
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              $1.t->set_inner( current_definition.top().type);
              declare($1);
              delete $1.name;
	      }
        ;

nomods_initdcl0:
          notype_declarator exception_specification_opt maybeasm maybe_attribute '='
          init { M::winternal("nomods_initdcl0 1"); }
/* Note how the declaration of the variable is in effect while its init is parsed! */
        | notype_declarator exception_specification_opt maybeasm maybe_attribute { M::winternal("nomods_initdcl0 2"); }
        ;

/* the * rules are dummies to accept the Apollo extended syntax
   so that the header files compile. */
maybe_attribute:
      /* empty */
        | attributes
        ;
 
attributes:
      attribute
        | attributes attribute
        ;

attribute:
      ATTRIBUTE '(' '(' attribute_list ')' ')' 
        ;

attribute_list:
      attrib
        | attribute_list ',' attrib 
        ;
 
attrib:
    /* empty */
        | any_word
        | any_word '(' IDENTIFIER ')'
        | any_word '(' IDENTIFIER ',' nonnull_exprlist ')'
        | any_word '(' nonnull_exprlist ')' { delete_ListTypes( $3 );}
        ;

/* This still leaves out most reserved keywords,
   shouldn't we include them?  */

any_word:
          identifier {}
        | SCSPEC {}
        | TYPESPEC {}
        | TYPE_QUAL {}
        ;

/* A nonempty list of identifiers, including typenames.  */
identifiers_or_typenames:
        identifier {}
        | identifiers_or_typenames ',' identifier
        ;

maybe_init:
        %prec EMPTY /* empty */ 
        | '=' init   {}
        ;
/*
 * nei sorgenti originali del GCC 2.7.2
 * ********************************* WARNING: maybe_init: manca il ';' alla
 * fine della regola!!!!!!!!!!!!!!!
 */



init:
          expr_no_commas %prec '=' { $$.begin= $1.begin; $$.end=$1.end; }
        | '{' '}'                  { $$.begin= $1.begin_token_string; $$.end=$2.end_token_string; }
        | '{' initlist '}'         { $$.begin= $1.begin_token_string; $$.end=$3.end_token_string; }
        | '{' initlist ',' '}'     { $$.begin= $1.begin_token_string; $$.end=$4.end_token_string; }
        | error                    {}
        ;

/* This chain is built in reverse order,
   and put in forward order where initlist is used.  */
initlist:
          init                {}
        | initlist ',' init
        /* These are for labeled elements.  */
        | '[' expr_no_commas ']' init      {}
        | initlist ',' CASE expr_no_commas ':' init
        | identifier ':' init  {}
        | initlist ',' identifier ':' init
        ;

structsp:
          ENUM identifier '{' 
                 {
                 Enumeration_Entity* e=(Enumeration_Entity*)&current_scope.add_Entity( new Enumeration_Entity($2.s)); // WARNING: we have to be check if it is possible
                 Entity_Container* ec=e->get_container();
		 if ( ec->cl == Entity::AggregateType ||
		      ec->cl == Entity::Namespace  )
		      {
		      Grouping_Entity* ate=(Grouping_Entity*) ec;
		      ate->add_AuxSpace(e);
		      }
                 current_scope.Push(*e); 
                 }
          enumlist maybecomma_warn '}'  
                 {
		 $$= (Enumeration_Entity*)current_scope.Get_Container(); 
		 current_scope.Pop();  
		 }
        | ENUM identifier '{' '}' 
	         {
		 Enumeration_Entity* e= (Enumeration_Entity*)&current_scope.add_Entity( new Enumeration_Entity($2.s));
                 Entity_Container* ec=e->get_container();
		 if ( ec->cl == Entity::AggregateType ||
		      ec->cl == Entity::Namespace  )
		      {
		      Grouping_Entity* ate=(Grouping_Entity*) ec;
		      ate->add_AuxSpace(e);
		      }
		 $$=e;
		 }
        | ENUM '{' 
                 { 
                 Enumeration_Entity* e=(Enumeration_Entity*)&current_scope.add_Entity( new Enumeration_Entity(NULL) ); // WARNING: we have to be check if it is possible
                 current_scope.Push(*e);
                 }
          enumlist maybecomma_warn '}' 
                 {
		 Enumeration_Entity* e= (Enumeration_Entity*)current_scope.Get_Container(); 
                 Entity_Container* ec=e->get_container();
		 if ( ec->cl == Entity::AggregateType ||
		      ec->cl == Entity::Namespace  )
		      {
		      Grouping_Entity* ate=(Grouping_Entity*) ec;
		      ate->add_AuxSpace(e);
		      }
		 current_scope.Pop(); 
		 $$=e;
		 }
        | ENUM '{' '}' 
	         {
		 Enumeration_Entity* e=(Enumeration_Entity*)&current_scope.add_Entity( new Enumeration_Entity(NULL) ); 
                 Entity_Container* ec=e->get_container();
		 if ( ec->cl == Entity::AggregateType ||
		      ec->cl == Entity::Namespace  )
		      {
		      Grouping_Entity* ate=(Grouping_Entity*) ec;
		      ate->add_AuxSpace(e);
		      }
		 $$= e;
		 }
        | ENUM identifier 
	         {
		 Enumeration_Entity* e=(Enumeration_Entity*)&current_scope.add_Entity( new Enumeration_Entity($2.s));
                 Entity_Container* ec=e->get_container();
		 if ( ec->cl == Entity::AggregateType ||
		      ec->cl == Entity::Namespace  )
		      {
		      Grouping_Entity* ate=(Grouping_Entity*) ec;
		      ate->add_AuxSpace(e);
		      }
		 $$= e;
		 }
        | ENUM complex_type_name {M::internal("structsp 1");}
        | TYPENAME_KEYWORD complex_type_name {M::internal("structsp 2");}
        /* C++ extensions, merged with C to avoid shift/reduce conflicts */
        | class_head left_curly
              {
	      current_scope.Push(*$1);
	      current_definition.push();
              }
          opt.component_decl_list '}' 
              {
              $$=$1;
	      modify_class($1, $2, $5 );  // scope is in the class
	      current_definition.pull();
              current_scope.Pop();
              }
        | class_head  %prec EMPTY  {$$=$1;}
        ;

maybecomma:
          /* empty */
        | ','  {}
        ;

maybecomma_warn:
          /* empty */
        | ','  {}
        ;

aggr:     AGGR
        | aggr SCSPEC
        | aggr TYPESPEC
        | aggr TYPE_QUAL
        | aggr AGGR
        ;

specialization:
          aggr template_type_name ';'
	    { 
	    L->put_back_token( $3 );
	    //M::winternal("specialization: not instantiated template.");
	    }
        ;

named_class_head_sans_basetype:
          aggr identifier 
             {
	     AggregateType_Entity* ae= current_scope.Name_LookUp_aggregate(Entity_Name($2.s));
             if (ae) 
                  $$= ae;  
             else   // if there isn't it value as a declaration then we add the entity
		  $$= (AggregateType_Entity*)&current_scope.add_Entity( new AggregateType_Entity($2.s) );
             }
        | specialization { 
	                 //M::winternal("named_class_head_sans_basetype: dangerous situation");
	                 }
        ;

named_class_head_sans_basetype_defn:
          aggr identifier_defn %prec EMPTY
                   {
                   AggregateType_Entity* c;
                   if ( current_definition.top().sc & Friend )   // is it declaring a friend?
                        {
                        M::warning(6, "ingnored friend");
                        }
                   else
                        {
			if ( current_scope.Get_Container()->cl== Entity::Template )   // are we in a template definition
			      { 
                              c=(AggregateType_Entity*)&current_scope.add_Entity( new AggregateType_Entity($2.s) ); // this is the class in the template
			      Entity_Container* templ= current_scope.Get_Container(); // this is the template
			      Entity_Container* templ_container= templ->get_container(); // this is the Entity_Container where the template is
                              templ_container->change_name( templ,  Entity_Name($2.s)); // update the name of the template
			      }
                        else	
			      {
			      AggregateType_Entity* ae;
			      if ( $2.e && $2.e->cl==Entity::AggregateType )    // it could be a non aggregate
				  ae= (AggregateType_Entity*)$2.e;
			      else
				  ae= current_scope.Name_LookUp_aggregate(Entity_Name($2.s));
				
			      // here we are sure that if ae!=NULL point to an AggregateType
			      
                              if ( ae )   // is there already an entity for this identifier?
                                 if ( current_scope.can_be_shadowed(ae) )  // e` definita in superficie?
                                     c= (AggregateType_Entity*)&current_scope.add_Entity( new AggregateType_Entity($2.s) ) ;
                                 else
                                     c= (AggregateType_Entity*)ae;   // here we don't complain about redefinition, but the compiler will do it
                              else 
                                 c= (AggregateType_Entity*)&current_scope.add_Entity( new AggregateType_Entity($2.s) );
                              $$= c;
			      }
                        }
                   }
        | aggr template_type_name '{' 
	   { current_skip_buffer= &Null_skip_buffer;} 
	  skip_till_right_brace 
	   {M::winternal("template user definition skipped 1");}
        | aggr template_type_name ':' 
	   { current_skip_buffer= &Null_skip_buffer;} 
	  skip_till_left_brace  skip_till_right_brace  
	   {M::winternal("template user definition skipped 2");}
        ;

named_complex_class_head_sans_basetype:
          aggr nested_name_specifier identifier {reset_type_qualifier_scope();}
        | aggr template_type %prec EMPTY        {reset_type_qualifier_scope();}
        ;

do_xref_defn: /* empty */ %prec EMPTY
        ;

named_class_head:
          named_class_head_sans_basetype %prec EMPTY {$$=$1;}
        | named_class_head_sans_basetype_defn do_xref_defn { current_scope.Push(*$$);}
          maybe_base_class_list maybe_extension_clauses %prec EMPTY
             {
	     current_scope.Pop();
	     $$=$1;
	     }
        | named_complex_class_head_sans_basetype maybe_base_class_list { M::internal("in named_class_head");}
        ;

maybe_extension_clauses:    /* this is added to handle the extend, extndible clauses */
        /*empty*/
        | EXTENSIBLE
             { 
	     ( (AggregateType_Entity*)current_scope.Get_Container() )->set_extensible();
	     delete_source( $1->begin_token_string, $1->end_token_string );
	     }
        | EXTEND complete_type_name
             {
	     if ( 
		    ! ( (AggregateType_Entity*)current_scope.Get_Container() )->
		                  add_extend((AggregateType_Entity*)$2.e) 
		)
                 M::warning( 1, "can't extend a 'non-extensible'" );
	     delete_source( $1->begin_token_string, $2.end_string );
	     }
	| DYNAMIC complete_type_name
             {
	     if ( 
		    ! ( (AggregateType_Entity*)current_scope.Get_Container() )->
		                  add_dynamic((AggregateType_Entity*)$2.e) 
		)
                 M::warning( 1, "can't dynamic-inherit a 'non-extensible'" );
	     delete_source( $1->begin_token_string, $2.end_string );
	     }
        | EXTENSIBLE EXTEND complete_type_name
             {
             ( (AggregateType_Entity*)current_scope.Get_Container() )->set_extensible();
	     if ( 
		    ! ( (AggregateType_Entity*)current_scope.Get_Container() )->
		                  add_extend((AggregateType_Entity*)$3.e) 
		)
                 M::warning( 1, "can't extend a 'non-extensible'" );
	     delete_source( $1->begin_token_string, $3.end_string );
	     }
        | EXTENSIBLE DYNAMIC complete_type_name
             {
             ( (AggregateType_Entity*)current_scope.Get_Container() )->set_extensible();
	     if ( 
		    ! ( (AggregateType_Entity*)current_scope.Get_Container() )->
		                  add_dynamic((AggregateType_Entity*)$3.e) 
		)
                 M::warning( 1, "can't dynamic-inherit a 'non-extensible'" );
	     delete_source( $1->begin_token_string, $3.end_string );
	     }
        ;

unnamed_class_head: 
        aggr '{'     
           {
           $$= (AggregateType_Entity*)&current_scope.add_Entity( new AggregateType_Entity(NULL) );
	   L->put_back_token( $2 );
	   }
        ;

class_head: 
          unnamed_class_head {$$=$1;}
        | named_class_head {$$=$1;}
        ;

maybe_base_class_list:
          %prec EMPTY /* empty */
        | ':' see_typename   %prec EMPTY { delete_source( $1->begin_token_string, $1->end_token_string); }
        | ':' see_typename base_class_list   %prec EMPTY {}
        ;

base_class_list:
          base_class
        | base_class_list ',' see_typename base_class
        ;

base_class:
          base_class.1      
        | base_class_access_list see_typename base_class.1
        ;

base_class.1:
          complete_type_name
             {
	     if ( $1.e->cl == Entity::AggregateType )
	         ( (AggregateType_Entity*)current_scope.Get_Container() )->
	                 add_inherit((AggregateType_Entity*)$1.e) ;
             else
                 M::warning( 1, "inherit from non class" );
             }
        | SIGOF '(' expr ')'    {M::internal("base_class.1 1");}
        | SIGOF '(' type_id ')' {M::internal("base_class.1 2");}
        ;

base_class_access_list:
          VISSPEC see_typename {}
        | SCSPEC see_typename {}
        | base_class_access_list VISSPEC see_typename {}
        | base_class_access_list SCSPEC see_typename {}
        ;

left_curly: '{' { $$= $1; }
        ;

opt.component_decl_list:
        /* empty */
        | component_decl_list
        | opt.component_decl_list VISSPEC ':' component_decl_list
        | opt.component_decl_list VISSPEC ':'
        ;

/* Note: we no longer warn about the semicolon after a component_decl_list.
   ARM $9.2 says that the semicolon is optional, and therefore allowed.  */
component_decl_list:
          component_decl
        | component_decl_list component_decl
        ;

component_decl:
          component_decl_1 ';'  { current_definition.top().null();}
        | component_decl_1 '}'  { current_definition.top().null();}
        /* C++: handle constructors, destructors and inline functions */
        /* note that INLINE is like a TYPESPEC */
        | fn.def2 ':' /* base_init compstmt */
	     { current_skip_buffer= &Null_skip_buffer;} 
          skip_till_left_brace
          skip_till_right_brace /* added my Maurizio Pizzonia*/
             {
	     modify_function( $1, $4, $5, true);
	     current_definition.top().null();
	     }
        | fn.def2 TRY /* base_init compstmt */
             { M::internal("component_decl 2"); }
        | fn.def2 RETURN /* base_init compstmt */
             { M::internal("component_decl 2"); }
        | fn.def2 '{' /* nodecls compstmt */
	     { current_skip_buffer= &Null_skip_buffer;} 
           skip_till_right_brace /* added my Maurizio Pizzonia*/
             { 
	     modify_function( $1, $2, $4, false);
	     current_definition.top().null();
	     }
        | ';'  { current_definition.top().null();}
        | EXTEND 
             {      // ECO extension
	     AggregateType_Entity* a=(AggregateType_Entity*)current_scope.Get_Container();
	     current_scope.Push( a->Extender_Methods );
	     } 
          fn.def2 ';' 
             { 
	     current_definition.top().null();
	     current_scope.Pop();
	     delete_source($1->begin_token_string, $4->end_token_string );
	     }
        ;

skip_till_right_brace:   // eat tokens until a right brace without a left one that must be parsed before
        /* empty */
              {             
              YYSTYPE value;
              t_id  id;
              unsigned level=1; // it is already in a block
              do  {
                  id= yylex(&value);           // eat at least one token
		  current_skip_buffer->add_token(value.UA_token); //save it in the current skip buffer
                  switch(id)
                     {
                     case '{': level++; break;
                     case '}': level--; break;
                     default: /*do nothing*/  break;
                     }
                  }
              while( level>0 && id!=END_OF_COMPILATION_UNIT );
	      $$= value.UA_token;
              yyclearin;
              }
        ;



skip_till_left_brace:   // eat tokens until a left brace (included)
        /* empty */
              {             
              YYSTYPE value;
              t_id  id;
              yyclearin;
              do  {
                  id= yylex(&value);           // eat at least one token
		  current_skip_buffer->add_token(value.UA_token); //save it in the current skip buffer
                  }
              while( id!='{' && id!=END_OF_COMPILATION_UNIT );
	      $$= value.UA_token;
              }
        ;


component_decl_1:
        /* Do not add a "typed_declspecs declarator" rule here for
           speed; we need to call grok_x_components for enums, so the
           speedup would be insignificant.  */
          typed_declspecs components {current_definition.top().null();}
        | declmods notype_components 
        | notype_declarator exception_specification_opt maybeasm maybe_attribute maybe_init 
              {
	      if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
              $1.t->set_inner( current_definition.top().type);
              declare($1);
              delete $1.name;
              }
        | ':' expr_no_commas {M::winternal("component_decl_1 4");}
        | error

        /* These rules introduce a reduce/reduce conflict; in
                typedef int foo, bar;
                class A {
                  foo (bar);
                };
           should "A::foo" be declared as a function or "A::bar" as a data
           member? In other words, is "bar" an after_type_declarator or a
           parmlist? */
        | typed_declspecs  begin_parmlist   parmlist ')' type_quals exception_specification_opt maybeasm maybe_attribute maybe_init 
	      { 
	      end_parameters();
	      if ( !current_definition.top().type ) M::internal("component_decl_1: no type setted for current definition");
	      if ( current_definition.top().type->k == Type::aggregate )
		  {
		  Aggregate* t= (Aggregate*)current_definition.top().type;
		  if ( t->TheAggregate->name == current_scope.Get_Container()->name )      // this is a constructor
		      {
	              A_name_entity_type x;
	              x.name= new Entity_Name; // no name
	              x.e=NULL;
	              x.t= $2; x.t->set_inner( t ); // type is function returning aggregate
	              declare( x );
		      delete x.name;
		      }
		  else
		      M::winternal("component_decl_1: non catched definition 2");
		  }
	      else
		  M::winternal("component_decl_1: non catched definition 1");
	      }
        | typed_declspecs LEFT_RIGHT type_quals exception_specification_opt maybeasm maybe_attribute maybe_init 
	      {
	      if ( !current_definition.top().type ) M::internal("component_decl_1: no type setted for current definition");
	      if ( current_definition.top().type->k == Type::aggregate )
		  {
		  Aggregate* t= (Aggregate*)current_definition.top().type;
		  if ( t->TheAggregate->name == current_scope.Get_Container()->name )      // this is a constructor
		      {
	              A_name_entity_type x;
	              x.name= new Entity_Name; // no name
	              x.e=NULL;
	              x.t= new Function(current_scope.Get_Container()); x.t->set_inner( t ); // type is function returning aggregate
	              declare( x );
		      delete x.name;
		      }
		  else
		      M::winternal("component_decl_1: non catched definition 3");
		  }
	      else
		  M::winternal("component_decl_1: non catched definition 4");
	      }
        | using_decl   {M::winternal("component_decl_1 3");}
        ;

begin_parmlist: 
        '(' { $$= begin_parameters();}
         ;




/* The case of exactly one component is handled directly by component_decl. */
/* ??? Huh? ^^^ */
components:
          /* empty: possibly anonymous */  {}
        | component_declarator0
        | components ',' component_declarator
        ;

notype_components:
          /* empty: possibly anonymous */  {}
        | notype_component_declarator0 
                 {
	         if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
                 $1.t->set_inner( current_definition.top().type );
                 declare($1); 
                 delete $1.name; 
                 }
        | notype_components ',' notype_component_declarator
                 {
	         if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
                 $3.t->set_inner( current_definition.top().type );
                 declare($3); 
                 delete $3.name; 
                 }
        ;

component_declarator0:
          after_type_component_declarator0
                 { 
	         if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
                 $1.t->set_inner( current_definition.top().type );
                 declare($1); 
                 delete $1.name; 
                 }
        | notype_component_declarator0
                 { 
	         if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
                 $1.t->set_inner( current_definition.top().type );
                 declare($1); 
                 delete $1.name; 
                 }
        ;

component_declarator:
          after_type_component_declarator
                 { 
	         if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
                 $1.t->set_inner( current_definition.top().type );
                 declare($1); 
                 delete $1.name; 
                 }
        | notype_component_declarator
                 { 
	         if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
                 $1.t->set_inner( current_definition.top().type );
                 declare($1);   
                 delete $1.name; 
                 }
        ;

after_type_component_declarator0:
          after_type_declarator exception_specification_opt maybeasm maybe_attribute maybe_init {$$.name=$1.name; $$.e=$1.e; $$.t=$1.t;}
        | TYPENAME ':' expr_no_commas maybe_attribute 
	         {
		 $$.name= new Entity_Name( $1->aux.id.s );
		 $$.e= $1->aux.id.e;
		 $$.t= new Direct;
	         }
        ;

notype_component_declarator0:
          notype_declarator exception_specification_opt maybeasm maybe_attribute maybe_init 
                 {
		 $$.name=$1.name;
		 $$.e=$1.e;
		 $$.t=$1.t;
		 }
        | IDENTIFIER ':' expr_no_commas maybe_attribute 
	         {
		 $$.name= new Entity_Name( $1->aux.id.s );
		 $$.e= $1->aux.id.e;
		 $$.t= new Direct;
		 }
        | ':' expr_no_commas maybe_attribute 
	         { 
		 $$.name= new Entity_Name();
		 $$.e= NULL;
		 $$.t= new Direct;
		 }
        ;

after_type_component_declarator:
          after_type_declarator exception_specification_opt maybeasm maybe_attribute maybe_init { $$.name=$1.name; $$.e=$1.e; $$.t=$1.t; }
        | TYPENAME ':' expr_no_commas maybe_attribute 
	         {
		 $$.name= new Entity_Name( $1->aux.id.s );
		 $$.e= $1->aux.id.e;
		 $$.t= new Direct;
	         }
        ;

notype_component_declarator:
          notype_declarator exception_specification_opt maybeasm maybe_attribute maybe_init 
                 {
		 $$.name=$1.name;
		 $$.e=$1.e;
		 $$.t=$1.t;
		 }
        | IDENTIFIER ':' expr_no_commas maybe_attribute 
	         {
		 $$.name= new Entity_Name( $1->aux.id.s );
		 $$.e= $1->aux.id.e;
		 $$.t= new Direct;
		 }
        | ':' expr_no_commas maybe_attribute 
	         { 
		 $$.name= new Entity_Name();
		 $$.e= NULL;
		 $$.t= new Direct;
		 }
        ;

/* We chain the enumerators in reverse order.
   Because of the way enums are built, the order is
   insignificant.  Take advantage of this fact.  */

enumlist:
          enumerator
        | enumlist ',' enumerator
        ;

enumerator:
          identifier {current_scope.add_Entity(new EnumElement_Entity($1.s));}
        | identifier '=' expr_no_commas {current_scope.add_Entity(new EnumElement_Entity($1.s));}
        ;

/* ANSI new-type-id (5.3.4) */
new_type_id:
          type_specifier_seq new_declarator 
	        { 
		$$.t= $2; 
	        if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
		$$.t->set_inner(current_definition.top().type); 
		M::winternal("new_type_id 1 - possible problem with substitution border");
		$$.begin= 0; $$.end= 0;
		}
        | type_specifier_seq %prec EMPTY 
	        { 
	        if (!current_definition.top().type) current_definition.top().type= new Built_in(Int);
		$$.t= current_definition.top().type; 
		M::winternal("new_type_id 2"); 
		$$.begin= 0; $$.end= 0;
		}
        /* GNU extension to allow arrays of arbitrary types with
           non-constant dimension.  */
        | '(' type_id ')' '[' expr ']' { $$.t= $2 ; $$.t->set_inner( new Array(0) ); $$.begin= $1.begin_token_string; $$.end= $6.end_token_string;}
        ;

type_quals:
          /* empty */ %prec EMPTY { $$= None_TYPE_QUAL;}
        | type_quals TYPE_QUAL  {(int)$$|=$2.aux.tq;}
        ;

nonempty_type_quals:
          TYPE_QUAL {$$= $1.aux.tq;}
        | nonempty_type_quals TYPE_QUAL {(int)$$|=$2.aux.tq;}
        ;

/* These rules must follow the rules for function declarations
   and component declarations.  That way, longer rules are preferred.  */

suspend_mom:
          /* empty */
           ;
           
            
/* An expression which will not live on the momentary obstack.  */
nonmomentary_expr:
        suspend_mom expr {delete $2.t;}
        ;






          
         
maybe_parmlist:  
          suspend_mom begin_parmlist nonnull_exprlist ')'
               {        // WARNING: 'begin_parmlist' is in place of '(' to avoid a strange behavior
		        // (this rule was completely ignored), We need to add the following line to 
		        // undo the creation of fictitious 'Function'
	        delete_ListTypes( $3 );
	       end_parameters(); delete $2; 
	       $$= new Direct; M::winternal("maybe_parmlist 1");
	       }
        | suspend_mom begin_parmlist parmlist ')'   { end_parameters(); $$=$2;}
        | suspend_mom LEFT_RIGHT               { $$= new Function(current_scope.Get_Container());}
        | suspend_mom '(' error ')'            {M::winternal("maybe_parmlist 3");}
        ;





/* A declarator that is allowed only after an explicit typespec.  */
/* may all be followed by prec '.' */
after_type_declarator:
          '*' nonempty_type_quals after_type_declarator  %prec UNARY { $$= $3; $$.t->set_inner(new Pointer); }
        | '&' nonempty_type_quals after_type_declarator  %prec UNARY { $$= $3; $$.t->set_inner(new Reference); }
        | '*' after_type_declarator  %prec UNARY {$$= $2; $$.t->set_inner( new Pointer);}
        | '&' after_type_declarator  %prec UNARY {$$= $2; $$.t->set_inner( new Reference);}
        | ptr_to_mem type_quals after_type_declarator {  M::winternal("after_type_declarator 1");}
        | direct_after_type_declarator {$$=$1;}
        ;

qualified_type_name:
          type_name %prec EMPTY { $$=$1;}
        | nested_type           { $$=$1; }
        ;

nested_type:
        nested_name_specifier type_name %prec EMPTY { $$=$2; reset_type_qualifier_scope();}
          /* WARNING:
	   * the following roule is to avoid parse error for qualified type
	   * which definition are not catched (inside templates)
	   *
	   * It will be deleted when templates will be instantiated
	   */
    /*
        | nested_name_specifier PSEUDO_TYPE 
		 {
		 reset_type_qualifier_scope();
		 
		 // define a fictitious class
		 AggregateType_Entity* ae= (AggregateType_Entity*)&current_scope.add_Entity( new AggregateType_Entity("$$internal_fictitious$$") );
		 
                 $$.e= ae;
		 $$.end_string= $2.end_token_string;
		 }
     */
        ;

direct_after_type_declarator:
          direct_after_type_declarator
              { current_scope.Push(*$$.scope); }
	  maybe_parmlist 
              { current_scope.Pop(); }
	  type_quals %prec '.'
	      { $$=$1; $$.t->set_inner( $3 ); }
        | direct_after_type_declarator '[' nonmomentary_expr ']'           {M::internal("direct_after_type_declarator 2");}
        | direct_after_type_declarator '[' ']'         {M::internal("direct_after_type_declarator 3");}
        | '(' after_type_declarator ')'                {M::internal("direct_after_type_declarator 4");}
        | nested_name_specifier type_name %prec EMPTY  { reset_type_qualifier_scope();M::internal("direct_after_type_declarator 5");}
        | type_name %prec EMPTY                        {
                                                       $$.t=new Direct;
                                                       $$.e=$1.e;
                                                       $$.name=new Entity_Name($1.e->name); // the only interesting.
                                                       $$.scope= current_scope.Get_Container();	  
                                                       }
        ;

/* A declarator allowed whether or not there has been
   an explicit typespec.  These cannot redeclare a typedef-name.  */

notype_declarator:
          '*' nonempty_type_quals notype_declarator  %prec UNARY { $$= $3; $$.t->set_inner(new Pointer);}
        | '&' nonempty_type_quals notype_declarator  %prec UNARY { $$= $3; $$.t->set_inner(new Reference); }
        | '*' notype_declarator  %prec UNARY { $$= $2; $$.t->set_inner( new Pointer);}
        | '&' notype_declarator  %prec UNARY { $$= $2; $$.t->set_inner( new Reference); }
        | ptr_to_mem type_quals notype_declarator { M::winternal("notype_declarator 1");}
        | direct_notype_declarator {$$=$1;}
        ;

complex_notype_declarator:
          '*' nonempty_type_quals notype_declarator  %prec UNARY {$$= $3; $$.t->set_inner( new Pointer);}
        | '&' nonempty_type_quals notype_declarator  %prec UNARY {$$= $3; $$.t->set_inner( new Reference);}
        | '*' complex_notype_declarator  %prec UNARY {$$= $2; $$.t->set_inner( new Pointer);}
        | '&' complex_notype_declarator  %prec UNARY {$$= $2; $$.t->set_inner( new Reference);}
        | ptr_to_mem type_quals notype_declarator {M::internal("complex_notype_declarator 1");}
        | complex_direct_notype_declarator {$$=$1;}
        ;


complex_direct_notype_declarator:
          direct_notype_declarator 
              {current_scope.Push(*$$.scope);}
          maybe_parmlist 
              {current_scope.Pop();} 
          type_quals  %prec '.'
               {
               $$=$1; $$.t->set_inner( $3 );
               }
        | '(' complex_notype_declarator ')' {$$=$2;}
        | direct_notype_declarator '[' nonmomentary_expr ']' { $$=$1; $$.t->set_inner( new Array(2000)); }
        | direct_notype_declarator '[' ']'                   { $$=$1; $$.t->set_inner( new Array(0)); }
        | notype_qualified_id 
                 { 
                 $$.name=$1.name;
                 $$.e=$1.e;
                 $$.scope= $1.scope;
                 $$.t=new Direct;
                 }
        ;

qualified_id:
        nested_name_specifier unqualified_id
            { 
            $$.name= $2.name;
            $$.e= $2.e;
            $$.scope= current_scope.Get_Container();
            reset_type_qualifier_scope();
	    $$.begin=$1.begin; $$.end=$2.end;
            }
        ;

notype_qualified_id:
        nested_name_specifier notype_unqualified_id
            { 
            $$.name= $2.name;
            $$.e= $2.e;
            $$.scope= current_scope.Get_Container();
            reset_type_qualifier_scope();
	    $$.begin=$1.begin; $$.end=$2.end;
            }
        ;

overqualified_id:
          notype_qualified_id              {reset_type_qualifier_scope(); $$=$1; $$.begin=$1.begin; $$.end=$1.end; }
        | global_scope notype_qualified_id {$$=$2; reset_type_qualifier_scope(); $$.begin=$1.begin; $$.end=$2.end;}
        ;

functional_cast:
          typespec '(' nonnull_exprlist ')'    { $$.t= current_definition.top().type;  delete_ListTypes( $3 ); $$.begin=0; $$.end=$4.end_token_string;}
        | typespec '(' expr_or_declarator ')'  { $$.t= current_definition.top().type;  $$.begin=0; $$.end=$4.end_token_string;}
        | typespec fcast_or_absdcl %prec EMPTY { $$.t= current_definition.top().type;  $$.begin=0; $$.end=$2.end;}
        ;

type_name:
          TYPENAME                  
              {
              reset_type_qualifier_scope();
              $$.e=$1->aux.id.e;
	      $$.end_string= $1.end_token_string;
              }
        | template_type %prec EMPTY 
              { 
              reset_type_qualifier_scope();
              $$=$1;
	      }
        ;

nested_name_specifier:
          nested_name_specifier_1 
              {
              if (scope_for_scope_qualifier)
                  current_scope.Change(*$1.c);
              else
                  {
                  current_scope.Push(*$1.c);
                  scope_for_scope_qualifier=true;
                  }
	      $$.begin=$1.begin; $$.end=$1.end;
              }
        | nested_name_specifier  nested_name_specifier_1
              {
              current_scope.Change(*$2.c); // scope_for_scope_qualifier==true for sure
	      $$.begin=$1.begin; $$.end=$2.end;
              }  
        ;

/* Why the @#$%^& do type_name and notype_identifier need to be expanded
   inline here?!?  (jason) */
nested_name_specifier_1:      /* return the Grouping_Entity pointer based on the current_scope*/
          TYPENAME SCOPE
              {
              //WARNING: here the entity is searched in the scope `Scope_Name_LookUp()' but this is
              //   right only for the first symbol.
              //   e.g.: `A::B::C'      
              //          `A' must be searched in global scope
              //          `B' must be searched locally to `A'
              //          `C' must be searched locally to  `B'
              //  this behave correctly for correct source.
              //
              // maybe this search is redaundant because the $1.aux.id.e is correctly set by the lexer
              // but I don't know if `current_scope' change since early `nested_name_specifier' reach this
              // token.
              //
              Entity* e= current_scope.Get_Container()->Scope_Name_LookUp(Entity_Name($1.aux.id.s));
              if ( e )
		     {
                     if ( e->cl== Entity::AggregateType )
                           $$.c=(Grouping_Entity*)e;
                     else if ( e->cl == Entity::Simple )  // surely it is a typedef (see behave of lexer)
		           {
			   Type* t= ((Simple_Entity*)e)->Get_Type();
			   if (t->k == Type::aggregate )
			       $$.c= ((Aggregate*)t)->TheAggregate;
			   else
			       {
			       M::warning(4, "accessed to a member for non aggregate type");
			       $$.c=current_scope.Get_Container();
			       }
			   }
		     else
                           {
                           M::winternal("nested_name_specifier_1: TYPENAME not found");
                           $$.c=current_scope.Get_Container();
                           }     
		     }
	      $$.begin=$1.begin_token_string; $$.end=$2.end_token_string;
              }
        | NSNAME SCOPE 
              {
              Entity* e= current_scope.Get_Container()->Scope_Name_LookUp(Entity_Name($1.aux.id.s));
              if ( e && e->cl== Entity::AggregateType ) 
                 $$.c=(Grouping_Entity*)e;
              else
                 {
                 M::winternal( "nested_name_specifier_1: NSNAME not found");
                 $$.c=current_scope.Get_Container();
                 }
	      $$.begin=$1.begin_token_string; $$.end=$2.end_token_string;
              }
        | template_type SCOPE 
              {
              //M::winternal("nested_name_specifier_1 - possible problems with substitution border");
              $$.c=(Template_Entity*)$1.e;
	      $$.begin=0; $$.end=$2.end_token_string;
              }
      /* 
        | IDENTIFIER SCOPE
              {
              Entity* e= current_scope.Get_Container()->Scope_Name_LookUp(Entity_Name($1.aux.id.s));
              if ( e && e->cl== Entity::AggregateType ) 
                 $$.c=(Grouping_Entity*)e;
              else
                 {
                 M::winternal("nested_name_specifier_1: IDENTIFIER not found");
                 $$.c=current_scope.Get_Container();
                 }
	      $$.begin=$1.begin_token_string; $$.end=$2.end_token_string;
              }
        | PTYPENAME SCOPE 
              {
              internal("nested_name_specifier_1 (PTYPENAME)");
              $$.c=current_scope.Get_Container();
	      $$.begin=$1.begin_token_string; $$.end=$2.end_token_string;
              }
       */
        ;

complete_type_name:
          qualified_type_name              { $$=$1; reset_type_qualifier_scope();}
        | global_scope qualified_type_name { $$=$2; reset_type_qualifier_scope(); M::winternal("complete_type_name");}
        ;

complex_type_name:
          nested_type                      { $$=$1;} 
        | global_scope qualified_type_name { $$=$2; reset_type_qualifier_scope();}
	;

ptr_to_mem:                        /* WARNING: non so se e` giusto mettere i seguenti reset */
          nested_name_specifier '*'          { reset_type_qualifier_scope(); M::internal("ptr_to_mem 1");}
        | global_scope nested_name_specifier '*' { reset_type_qualifier_scope(); M::internal("ptr_to_mem 2");}
        ;

/* All uses of explicit global scope must go through this nonterminal so
   that got_scope will be set before yylex is called to get the next token. */
global_scope:
          SCOPE 
             {
             if (scope_for_scope_qualifier)
                  {
                  M::winternal("global_scope: there is already scope for type qualifiers on the stack" );
                  current_scope.Change(*Global);
                  }
             else
                  {
                  current_scope.Push(*Global);
                  scope_for_scope_qualifier=true;
                  }
	     $$.begin= $1.begin_token_string; $$.end= $1.end_token_string;
             }
        ;

/* ANSI new-declarator (5.3.4) */
new_declarator:
          '*' type_quals new_declarator { $$= $3; $$->set_inner( new Pointer );}
        | '*' type_quals  %prec EMPTY { $$= new Pointer; }
        | '&' type_quals new_declarator %prec EMPTY { $$= $3; $$->set_inner( new Reference );}
        | '&' type_quals %prec EMPTY  { $$= new Reference; }
        | ptr_to_mem type_quals %prec EMPTY    {M::internal("new_declarator 1");}
        | ptr_to_mem type_quals new_declarator {M::internal("new_declarator 2");}
        | direct_new_declarator %prec EMPTY {$$=$1;}
        ;

/* ANSI direct-new-declarator (5.3.4) */
direct_new_declarator:
          '[' expr ']'   { delete $2.t; $$=new Array(0); }
        | direct_new_declarator '[' nonmomentary_expr ']' { $$=new Array(0); $$->set_inner( $1 ); }
        ;

/* ANSI abstract-declarator (8.1) */
absdcl:
          '*' nonempty_type_quals absdcl       { $$= $3; $$->set_inner( new Pointer ); }
        | '*' absdcl                           { $$= $2; $$->set_inner( new Pointer ); }
        | '*' nonempty_type_quals  %prec EMPTY { $$= new Direct; $$->set_inner( new Pointer ); }
        | '*' %prec EMPTY                      { $$= new Direct; $$->set_inner( new Pointer ); }
        | '&' nonempty_type_quals absdcl       { $$= $3; $$->set_inner( new Reference ); }
        | '&' absdcl                           { $$= $2; $$->set_inner( new Reference ); }
        | '&' nonempty_type_quals %prec EMPTY  { $$= new Direct; $$->set_inner( new Reference ); }
        | '&' %prec EMPTY                      { $$= new Direct; $$->set_inner( new Reference ); }
        | ptr_to_mem type_quals %prec EMPTY    { M::internal("absdcl 1");}
        | ptr_to_mem type_quals absdcl         { M::internal("absdcl 2");}
        | direct_abstract_declarator %prec EMPTY { $$=$1; }
        ;

/* ANSI direct-abstract-declarator (8.1) */
direct_abstract_declarator:
          '(' absdcl ')'                                                  {$$=$2;}
          /* `(typedef)1' is `int'.  */
        | PAREN_STAR_PAREN                                                { $$= new Pointer;}
        | direct_abstract_declarator begin_parmlist parmlist ')' type_quals  %prec '.' {end_parameters(); $$=$1; $$->set_inner($2); }
        | direct_abstract_declarator LEFT_RIGHT type_quals  %prec '.'       { $$=$1; $$->set_inner( new Function( current_scope.Get_Container()) ); }
        | direct_abstract_declarator '[' nonmomentary_expr ']'  %prec '.'   { $$=$1; $$->set_inner( new Array( 2001)); }
        | direct_abstract_declarator '[' ']'  %prec '.'                     { $$=$1; $$->set_inner( new Array( 0)); }
        | '(' complex_parmlist ')' type_quals  %prec '.'           {M::internal("direct_abstract_declarator 2");}
        | regcast_or_absdcl type_quals %prec '.'                   {M::internal("direct_abstract_declarator 3");}
        | fcast_or_absdcl type_quals %prec '.'                     {M::internal("direct_abstract_declarator 4");}
        | '[' nonmomentary_expr ']'  %prec '.'                     { $$= new Array(  2001); }
        | '[' ']'  %prec '.'                                       { $$= new Array(  0); }
        ;

/* For C++, decls and stmts can be intermixed, so we don't need to
   have a special rule that won't start parsing the stmt section
   until we have a stmt that parses without errors.  */

stmts:
          stmt
        | errstmt
        | stmts stmt
        | stmts errstmt
        ;

errstmt:  error ';'  { current_definition.top().null();}
        ;

/* build the LET_STMT node before parsing its contents,
  so that any LET_STMTs within the context can have their display pointers
  set up to point at this one.  */

.pushlevel:  /* empty */
	 {
	 Namespace_Entity* n= new Namespace_Entity(NULL);
	 current_scope.add_Entity( n );
	 current_scope.Push(*n);
	 }
        ;

.poplevel:   /* empty */
	 {  current_scope.Pop(); }
        ;

/* Read zero or more forward-declarations for labels
   that nested functions can jump to.  */
maybe_label_decls:
          /* empty */
        | label_decls
        ;

label_decls:
          label_decl
        | label_decls label_decl
        ;

label_decl:
          LABEL identifiers_or_typenames ';'
        ;

/* This is the body of a function definition.
   It causes syntax errors to ignore to the next openbrace.  */
compstmt_or_error:
          compstmt
        | error compstmt
        ;

compstmt: '{' .pushlevel compstmtend .poplevel {}
        ;

simple_if:
          IF
          .pushlevel paren_cond_or_null
          implicitly_scoped_stmt
        ;

implicitly_scoped_stmt:
          compstmt
        | .pushlevel simple_stmt .poplevel
        ;

stmt:
          compstmt
        | simple_stmt
        ;

simple_stmt:
          decl
        | expr ';' { delete $1.t; }
        | simple_if ELSE
          implicitly_scoped_stmt
          .poplevel
        | simple_if %prec IF 
	       {  // this is like .poplevel but avoid a shift/reduce conflict
	       current_scope.Pop(); 
	       }
        | WHILE
          .pushlevel paren_cond_or_null
          already_scoped_stmt .poplevel
        | DO
          implicitly_scoped_stmt WHILE
          paren_expr_or_null ';'
        | FOR .pushlevel
          '(' for.init.statement
          .pushlevel xcond ';'
          xexpr ')'
          already_scoped_stmt .poplevel .poplevel
        | SWITCH .pushlevel '(' condition ')'
          implicitly_scoped_stmt
          .poplevel
        | CASE expr_no_commas ':'
          stmt
        | CASE expr_no_commas ELLIPSIS expr_no_commas ':'
          stmt
        | DEFAULT ':'
          stmt
        | BREAK ';'
        | CONTINUE ';'
        | RETURN ';'       {}
        | RETURN expr ';'  {}
        | asm_keyword maybe_type_qual '(' string ')' ';'
        /* This is the case with just output operands.  */
        | asm_keyword maybe_type_qual '(' string ':' asm_operands ')' ';'
        /* This is the case with input operands as well.  */
        | asm_keyword maybe_type_qual '(' string ':' asm_operands ':' asm_operands ')' ';'
        /* This is the case with clobbered registers as well.  */
        | asm_keyword maybe_type_qual '(' string ':' asm_operands ':'
          asm_operands ':' asm_clobbers ')' ';'
        | GOTO '*' expr ';'
        | GOTO identifier ';'
        | label_colon stmt
        | label_colon '}'
        | ';' {}
        | try_block
        ;

function_try_block:
          TRY
          ctor_initializer_opt compstmt_or_error
          handler_seq
        ;

try_block:
          TRY
          compstmt
          handler_seq
        ;

handler_seq:
          /* empty */
        | handler_seq CATCH .pushlevel
          handler_args
          compstmt
          .poplevel
        ;

type_specifier_seq:
          typed_typespecs %prec EMPTY  
        | nonempty_type_quals %prec EMPTY {(int)current_definition.top().tq|=$1;}
        ;

handler_args:
          '(' ELLIPSIS ')'  {}
			       
        /* This doesn't allow reference parameters, the below does.
        | '(' type_specifier_seq absdcl ')'
        | '(' type_specifier_seq ')'
        | '(' type_specifier_seq notype_declarator ')'
        | '(' typed_typespecs after_type_declarator ')'  */
			     
        /*This allows reference parameters... */          
        | '(' parm ')'    {}                               
        ;

label_colon:
          IDENTIFIER ':' {}
        | PTYPENAME ':' {}
        | TYPENAME ':' {}
        ;

for.init.statement:
          xexpr ';'
        | decl
        | '{' compstmtend {}
        ;

/* Either a type-qualifier or nothing.  First thing in an `asm' statement.  */

maybe_type_qual:
        /* empty */
        | TYPE_QUAL {}
        ;

xexpr:
        /* empty */
        | expr   { delete $1.t;}
        | error
        ;

/* These are the operands other than the first string and colon
   in  asm ("addextend %2,%1": "=dm" (x), "0" (y), "g" (*x))  */
asm_operands: /* empty */
        | nonnull_asm_operands
        ;

nonnull_asm_operands:
          asm_operand
        | nonnull_asm_operands ',' asm_operand
        ;

asm_operand:
          STRING '(' expr ')'  {}
        ;

asm_clobbers:
          STRING                   {}
        | asm_clobbers ',' STRING  {}
        ;

/* This is what appears inside the parens in a function declarator.
   Its value is represented in the format that grokdeclarator expects.

   In C++, declaring a function with no parameters
   means that that function takes *no* parameters.  */

parmlist:  
        /* empty */
        | complex_parmlist  
        | type_id   { current_scope.add_Entity( new Simple_Entity(NULL, None_SCSPEC, $1 ) ); }
        ;
 
/* This nonterminal does not include the common sequence '(' type_id ')',
   as it is ambiguous and must be disambiguated elsewhere.  */

complex_parmlist:
          parms                        
        | parms_comma ELLIPSIS         {M::winternal("complex_paramlist 1");}
        /* C++ allows an ellipsis without a separating ',' */
        | parms ELLIPSIS               {M::winternal("complex_paramlist 2");}
        | type_id ELLIPSIS             {M::winternal("complex_paramlist 3");}
        | ELLIPSIS                     {M::winternal("complex_paramlist 4");}
        | TYPENAME_ELLIPSIS            {M::winternal("complex_paramlist 5");}
        | parms TYPENAME_ELLIPSIS      {M::winternal("complex_paramlist 6");}
        | type_id TYPENAME_ELLIPSIS    {M::winternal("complex_paramlist 7");}
        | parms ':'                    {M::internal("complex_paramlist 8");}
        | type_id ':'                  {M::internal("complex_paramlist 9");}
        ;

/* A nonempty list of parameter declarations or type names.  */
parms:
          named_parm             
        | parm '=' init                  
        | parms_comma full_parm          
        | parms_comma bad_parm           
        | parms_comma bad_parm '=' init  {M::internal("parms 4");}
        ;

parms_comma:
          parms ','   { }
        | type_id ',' { current_scope.add_Entity( new Simple_Entity(NULL, None_SCSPEC, $1 ) ); }
        ;

/* A single parameter declaration or parameter type name,
   as found in a parmlist.  The first four cases make up for 10%
   of the time spent parsing C++.  We cannot use them because
   of `int id[]' which won't get parsed properly.  */
named_parm:
/*
          typed_declspecs dont_see_typename '*' IDENTIFIER
        | typed_declspecs dont_see_typename '&' IDENTIFIER
        | TYPENAME IDENTIFIER
        | TYPESPEC IDENTIFIER
        | */
        
        /* Here we expand typed_declspecs inline to avoid mis-parsing of
           TYPESPEC IDENTIFIER.  */
          typed_declspecs1  declarator 
        | typed_typespecs  declarator  
        | typespec declarator 
        | typed_declspecs1 absdcl     {/*M::internal("named_parm 3");*/}
        | typed_declspecs1 %prec EMPTY {/*M::internal("named_parm 4");*/}
        | declmods notype_declarator  
        ;


full_parm:
          parm maybe_init {}
        ;

parm:
        named_parm { }
        | type_id { current_scope.add_Entity( new Simple_Entity(NULL, None_SCSPEC, $1 ) ); }
        ;

see_typename: %prec EMPTY
        ;

/* 
dont_see_typename: %prec EMPTY
        ; 

try_for_typename:
        ;
*/

bad_parm:
          /* empty */ %prec EMPTY
        | notype_declarator {}
        ;

exception_specification_opt:
          %prec EMPTY /* empty */
        | THROW '(' ansi_raise_identifiers  ')' %prec EMPTY
        | THROW LEFT_RIGHT %prec EMPTY
        ;

ansi_raise_identifier:
          type_id     { M::internal("ansi_raise_identifier");}
        ;

ansi_raise_identifiers:
          ansi_raise_identifier
        | ansi_raise_identifiers ',' ansi_raise_identifier
        ;

conversion_declarator:   /* used only in operator_name */
          /* empty */ %prec EMPTY    {$$=new Direct;}
        | '*' type_quals conversion_declarator {$$=$3; $$->set_inner( new Pointer );}
        | '&' type_quals conversion_declarator {$$=$3; $$->set_inner( new Reference );}
        | ptr_to_mem type_quals conversion_declarator { M::internal("conversion_declarator");}
        ;

operator: OPERATOR { $$.begin= $1.begin_token_string; $$.end= $1.end_token_string; }
        ;

operator_name:
          operator '*' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Mul);}
        | operator '/' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Div);}
        | operator '%' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Mod);}
        | operator '+' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Sum);}
        | operator '-' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Sub);}
        | operator '&' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Bit_And);}
        | operator '|' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Bit_Or);}
        | operator '^' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Bit_Xor);}
        | operator '~' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Bit_Not);}
        | operator ',' {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Comma);}
        | operator ARITHCOMPARE         {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier($2->aux.op);}
        | operator '<'                  {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Less);}
        | operator '>'                  {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Greater);}
        | operator EQCOMPARE            {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier($2->aux.op);}
        | operator ASSIGN               {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier($2->aux.op);}
        | operator '='                  {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Assign);}
        | operator LSHIFT               {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(LShift);}
        | operator RSHIFT               {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(RShift);}
        | operator PLUSPLUS             {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(PlusPlus);}
        | operator MINUSMINUS           {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(MinusMinus);}
        | operator ANDAND               {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Logic_And);}
        | operator OROR                 {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Logic_Or);}
        | operator '!'                  {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Logic_Not);}
        | operator '?' ':'              {$$.begin= $1.begin; $$.end=$3.end_token_string; $$.oi= new Operator_Identifier(Question_Colon);}
        | operator MIN_MAX              {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(MinMax);}
        | operator POINTSAT  %prec EMPTY{$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Pointsat);}
        | operator POINTSAT_STAR  %prec EMPTY  {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Pointsat_Star);}
        | operator LEFT_RIGHT           {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Call);}
        | operator '[' ']'              {$$.begin= $1.begin; $$.end=$3.end_token_string; $$.oi= new Operator_Identifier( Array_deref);}
        | operator NEW %prec EMPTY      {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(New);}
        | operator DELETE %prec EMPTY   {$$.begin= $1.begin; $$.end=$2.end_token_string; $$.oi= new Operator_Identifier(Delete);}
        | operator NEW '[' ']'          {$$.begin= $1.begin; $$.end=$3.end_token_string; $$.oi= new Operator_Identifier(New_Array);}
        | operator DELETE '[' ']'       {$$.begin= $1.begin; $$.end=$3.end_token_string; $$.oi= new Operator_Identifier(Delete_Array);}
        /* Names here should be looked up in class scope ALSO.  */
        | operator type_specifier_seq  conversion_declarator {$$.begin= $1.begin; $$.end=0; $3->set_inner(current_definition.top().type); $$.oi= new Operator_Identifier($3); }
        | operator error {}
        ;


