
#ifndef __ENTITY_H__
#define __ENTITY_H__

#include <string.h>
#include <iostream.h>

#include <list.h>
#include <set.h>
//#include <deque.h>

#include <multiset.h>
#include "global.h"
#include "messages.h"

class Operator_Identifier;
class Entity_Name;

class Entity;
class Simple_Entity;
class Overloadable_Entity;
class EnumElement_Entity;


class Entity_Container;
class Grouping_Entity;
class Enumeration_Entity;
class AggregateType_Entity;
class Namespace_Entity;
class Parameters_Entity;
class Template_Entity;
class Type;


class Scope;

#include "parser_types.h"
#include "Window_on_a_Stream.h"



// ------------------------------------------------------- 
// this is a common interface to a source of token:
//
// and a minimal suppor for put_back_token

class Token_Source
    {
    private:
       A_token filter();
    protected:
       A_token wonted_token;    // if wonted_token.tokenid==0 then non wonted token is setted
                                //    else it is returned by next token before any other token in the stream
             
       Window_on_a_Stream<A_token,10, Token_Source>* w; // this manage a window on the token stream
             
    public:
       Scope& current_scope;            // scope nel quale vengono cercati i simboli per vedere se sono tipi
       Token_Source(Scope&);
       void Init_Window(Window_on_a_Stream<A_token,10, Token_Source>* w);
                              // this must be called before calling next_token()
                              // 'w' can't be initialized in the consturctor because
                              // 'w' read some token at the initialization and 
                              // inner_lexer is pure virtual here
      
    
    
       A_token next_token();
       virtual A_token inner_lexer()=0;
       A_token inner_lexer1();
    
       virtual void put_back_token( const A_token& );    // this has the effect to set the next wonted token
    };



// ------------------------------------------------------ Token_Buffer
// common interface to every buffer of token
//
class Token_Buffer: public Token_Source
    {
    public:
       virtual void add_token( const A_token& ) =0;
       Token_Buffer(Scope& sc);
       virtual A_token inner_lexer()=0;
    };

class Null_Token_Buffer: public Token_Buffer
    {
    public:
       Null_Token_Buffer();
       virtual void add_token( const A_token& );
       virtual A_token inner_lexer();
    };
// ------------------------------------------------------ Token_Buffer_tiny
//
// This is a container to store a sequence of tokens
// 
//  information on exact position of token string are omitted to minimize
//  memory usage. 
//  
//  when tokens are retrieved the following member are set to zero
//    begin_pre_string;  
//    begin_token_string;
//    end_token_string;  
//
//  It is ideal for store token sequence for templates to be parserd during
//  template instantiation
//

class Token_Buffer_tiny: public Token_Buffer
{
    private:
       struct tiny_token    // this is a version of A_token without some unimportant information
            {
            t_id tokenid;
	    
                                                                             // position in the CPP output
            unsigned line;                // line in which is the token (in the preprocessed file)
	      
                                                                              // position in the CPP input
            const char* include_file;     // file from which the token comes
            unsigned    include_file_line;// line of the file above
            
            Value aux;   // an auxiliar value for the identification of token (actualy only operator)  related to multiple string
	    };

       list<tiny_token> seq;   // here is stored the sequence of tokens
       typedef list<tiny_token>::iterator Iter;

       Iter position; // this is the current position

    public:
       Token_Buffer_tiny(Scope&);
       
       virtual void add_token( const A_token& );
       unsigned size() const { return seq.size(); }
       void reset();
       virtual A_token inner_lexer();
    };




// ---------------------------------------------- Type
class Elementary;
class Type
   {
   private:
      virtual void output_compound(ostream &os, const char* s) const;
   public:
      enum Type_kind
         {
	 direct,         // this is a type that beaves as if there isn't
         pointer,
         reference,
         array,
         aggregate,
         built_in,
         unresolved,
	 function   
         };
      Type_kind k;
   protected:
      Type(Type_kind f):
          k(f)
          {}
   public:
      void output(ostream &os) const;
      void output(ostream &os, const char* s) const;
      virtual ~Type() {}
      virtual void set_inner(Type* t)=0;
      virtual const Elementary* get_inner() const =0;           // return the ultimate type discarding Pointer, Reference, Array, Function; it can be Built_in or Aggregate
      virtual bool is_equal(const Type*)const =0;
      virtual void output_compound_pre(ostream &os, const char* s) const=0;
      virtual void output_compound_post(ostream &os) const=0;
      virtual Type* clone() const=0;
   };


class Direct_Compound: public Type
   {
   public:
      Type * inner_type;
      Direct_Compound(Type_kind t):
          Type(t),
          inner_type(NULL)
          {}
      ~Direct_Compound() 
          { delete inner_type; }
      virtual const Elementary* get_inner() const;
      virtual void set_inner(Type* t);
      virtual bool is_equal(const Type* t)const;
      virtual void output_compound_pre(ostream &os, const char* s) const;
      virtual void output_compound_post(ostream &os) const;
//      virtual bool operator<(const Type& ) const;
      virtual Type* clone() const =0;
   };

class Elementary: public Type
   {
   public:
      Elementary(Type_kind f):
         Type(f)
         {}
      virtual void set_inner(Type* t)
         { M::hl_internal(M::rec, "Elementary::set_inner(Type* t) called"); }
     const Elementary* get_inner() const 
         { return this; }
      virtual void out_Elementary(ostream& os) const =0;
      virtual void output_compound_pre(ostream &os, const char* s) const
         { os << s;};
      virtual void output_compound_post(ostream &os) const 
         { }
      virtual Type* clone() const =0;
   };

class Direct: public Direct_Compound
   {
   public:
      virtual void output_compound_pre(ostream &os, const char* s) const;
      virtual void output_compound_post(ostream &os) const;
   public:
      Direct():
         Direct_Compound(direct)
         {}
      virtual Type* clone() const;
   };


class Pointer: public Direct_Compound
   {
   public:
      virtual void output_compound_pre(ostream &os, const char* s) const;
      virtual void output_compound_post(ostream &os) const;
   public:
      Pointer():
         Direct_Compound(pointer)
         {}
      virtual Type* clone() const;
   };

class Reference: public Direct_Compound
   {
   public:
      virtual void output_compound_pre(ostream &os, const char* s) const;
      virtual void output_compound_post(ostream &os) const;
   
   public:
      Reference():
         Direct_Compound(reference)
         {}
      virtual Type* clone() const;
   };

class Entity_Container;
class Function: public Direct_Compound
   {
   public:
      virtual void output_compound_post(ostream &os) const;
   public:
      Parameters_Entity& Parameters;
      Function(Entity_Container*);
      ~Function();
      virtual bool is_equal(const Type* t)const;
      virtual Type* clone() const;
   };




class Array: public Direct_Compound
   {
   public:
      virtual void output_compound_post(ostream &os) const;
   public:
      const unsigned long size;
      Array( unsigned long n):
         Direct_Compound(array),
         size(n)
         {}
      virtual bool is_equal(const Type* t)const;
      virtual Type* clone() const;
   };



class Aggregate: public Elementary
   {
   public:
      Grouping_Entity *const TheAggregate; // reference (not owned)
      Aggregate(Grouping_Entity* e);
      void out_Elementary(ostream& os) const;
      virtual bool is_equal(const Type* t)const;
      virtual Type* clone() const;
   };


class Built_in: public Elementary
   {
   public:   
      TYPESPEC_value The_built_in;
      Built_in(TYPESPEC_value b):
         Elementary(built_in),
         The_built_in(b)
         {}
      ~Built_in() {}
      void out_Elementary(ostream& os) const;
      virtual bool is_equal(const Type* t)const;
      virtual Type* clone() const;
   };

class Unresolved_overloadable: public Elementary
   {
      Overloadable_Entity* ov;
   public:   
      Unresolved_overloadable(Overloadable_Entity* x):
         Elementary(unresolved),
         ov(x)
         {}
      void out_Elementary(ostream& os) const;
      virtual bool is_equal(const Type* t)const;
      virtual Type* clone() const;
      Type* resolve( const list<Type*>* ) const;
   };
// -------------------------------------------------------- Operator_Identifier






class Operator_Identifier
        {
        friend ostream& operator << ( ostream&, const Operator_Identifier&);
	public:
	     enum Kind 
	           {
		   conversion,
		   proper_operator,
		   null
		   };
	     Kind k;
	     Operator_ID  id;
	     Type * t;        // delete owning (but managment is hard), it could be better a reference (it is for now, no destructor)
	
	     Operator_Identifier( Operator_ID op):
	           k(proper_operator),
	           t(NULL),
	           id(op)
	           {}
	
	     Operator_Identifier( Type* q):     // it keep posses of q
	           k(conversion),
	           t(q),
	           id(Op_None)
	           {}
	     Operator_Identifier():
	           k(null),
	           t(NULL)
	           {}
	     bool operator==( const Operator_Identifier& oi ) const;
	     Operator_Identifier& operator= ( const Operator_Identifier& en )
	          {
		  k=en.k;
		  id=en.id;
		  t=en.t; // still
		  }
	     bool operator<(const Operator_Identifier& op ) const;

	};
ostream& operator << ( ostream&, const Operator_Identifier&);



class Entity_Name
        {
	friend ostream& operator << ( ostream&, const Entity_Name& );
	private:
	     Operator_Identifier op_id;
	     const char * str;                  // not owned, is a reference to the symbol repository
	                           
	                            //they has meaning only if k==template instance
             Template_Entity * Templ;   // reference 
	     list<Type*>* parameters;   // delete-owning (unless operator= that still it)
	public:
	     enum Kind
	            {
		    string,
		    constructor,
		    destructor,
		    template_instance,
		    oper,
		    none
		    }; 
	     Kind k;
	     Entity_Name(Operator_Identifier o):
	            k(oper),
	            op_id(o),
	            str(NULL)
	            {}
	     Entity_Name( const char * s):
	            k(string),
	            str(s)
	            {
		    if (!s) ( *(Kind*) &k )= none;
		    }
	     Entity_Name():
	            k(none),
	            str(NULL)
	            {}
	     Entity_Name( Kind kk):
	            str(NULL),
	            k(kk)
	            {
		    if (k != constructor )
		       M::hl_internal(M::fatal, "Entity_Name::Entity_Name(Kind): not a constructor");
		    }
	     Entity_Name( Kind kk, const char * s):
	            str(s),
	            k(kk)
	            {
		    if (k != destructor )
		       M::hl_internal(M::fatal, "Entity_Name::Entity_Name(Kind, const char*): not a destructor");
		    }
	     Entity_Name( Template_Entity* T, list<Type*>* parms):
	            str(NULL),
	            k(template_instance),
	            Templ(T),
	            parameters(parms)
	            { }
	      
	     Operator_Identifier Get_op() const { return op_id; }
	     const char * Get_str() const { return str; }
	     Template_Entity* Get_Template() const { return Templ; }
             list<Type*>*     Get_Template_parmeters() const { return parameters; }
	     bool operator==( const Entity_Name& en ) const;
	     Entity_Name& operator= ( const Entity_Name& en )
	          {
		  k= en.k;
		  op_id=en.op_id;
		  str=en.str;
		  Templ=en.Templ;
		  parameters= en.parameters;   // still
		  }
	     bool operator<( const Entity_Name& )const;
	     void check_consistency()const;
	};
ostream& operator << ( ostream&, const Entity_Name& );


// -------------------------------------------------------- Entity hierarchy

// These are classes to represent every C++ "thing".
// All these I call the Entity_DB and is in his minimal form an  instance of
//    Namespace_Entity but can contain a lot of things.
// It is managed in an incremental way.
// Client can perform only adding operation, no delete and no modify.
// It is possible to fill the blanks of not completely specified Entity.


class Grouping_Entity;
class Entity_Container;

class Entity
   {
   private:
      Entity( const Entity& );
      Entity& operator= ( const Entity& );
      Entity_Container * container;
   public:
      enum Entity_Class
         {
	 Unknown,
	 Simple,
	 Overloadable,
         Particular_overloadable_E,
         Parameters,
	 Function_E,
	 Template,
	 EnumElement,
	 Enumeration,
	 AggregateType,
         Namespace,
	 Extender_Methods_Container
	 };
      const Entity_Class cl;
      Entity_Name name;
      virtual void set_container( Entity_Container* ec)
            {
	    if(container)  M::hl_internal( M::rec, "Entity::set_container(): called more then one time" );
	    container= ec;
	    }
      Entity_Container* get_container() const 
            {return container;}
      const char * qualified_name() const;   // allocate a null terminated string containing the qualified name (starting form global namespace)
        // nuove funzioni di output
      virtual void simple_output(ostream &os, int indent) const =0;
      virtual void deep_output(ostream &os, int indent) const 
	  {simple_output(os,indent); };

    protected:
      Entity(const char *n, Entity_Class c=Unknown ):
         name(n),
         cl(c),
         container(NULL)
         {}
      Entity(const Entity_Name &n, Entity_Class c=Unknown ):
         name(n),
         cl(c),
         container(NULL)
         {}
      virtual ~Entity() {};
   };



class Simple_Entity: public Entity
   {
   protected:
      Type* TheType;   // NULL -> Type unspecified WARNING: proprio?
      SCSPEC_value TheSC;
   public:
      Simple_Entity(const char *n, SCSPEC_value sc, Type* t):
         Entity(n,Simple),
         TheType(t),
         TheSC(sc)
         {}
      Simple_Entity(const Entity_Name&n, SCSPEC_value sc, Type* t):
         Entity(n,Simple),
         TheType(t),
         TheSC(sc)
         { }
      void Set_Type( Type* t ) {TheType=t;}
      Type* Get_Type() const;
       
      SCSPEC_value Get_SC() const 
         { return TheSC; }
      virtual void simple_output(ostream &os, int indent) const;
   };






class Entity_Container: public Entity
   {
   private:
   protected:
      Entity_Container(const char *n, Entity_Class c);
      Entity_Container(const Entity_Name &, Entity_Class );
      virtual ~Entity_Container();
 
   public:
      virtual void simple_output(ostream &os, int indent) const=0;
      virtual void scope_output(ostream &os, int indent) const;
      virtual void deep_output(ostream &os, int indent) const;
      virtual void local_output(ostream &os, int indent) const;

      virtual Entity& add_Entity( Entity* e )
           {
	   M::hl_internal(M::rec, "Entity_Container::add_Entity() called");
	   }
   
                     // search only in this, search fail by defoult      
      virtual Entity*               Local_Name_LookUp(const Entity_Name& en) const;                         // search every but AggregateType_Entity
      virtual AggregateType_Entity* Local_Name_LookUp_aggregate(const Entity_Name& en) const; // search only for AggregateType_Entity
     
                     // search in this by Local_Name_LookUp  and in the container, too.
      virtual Entity*               Scope_Name_LookUp(const Entity_Name& en) const;            // search for all entity
      virtual AggregateType_Entity* Scope_Name_LookUp_aggregate(const Entity_Name& en) const; // search for every but AggregateType_Entity
     
      bool is_here( Entity * e )
         {return e->get_container()== this; }
      virtual void change_name( Entity*, const Entity_Name& )
           {
	   M::hl_internal(M::rec, "Entity_Container::change_name() called");
	   }
   };




   




class Grouping_Entity: public Entity_Container
   {
   private:
   protected:
      Grouping_Entity(const char *n, Entity_Class c);
      Grouping_Entity(const Entity_Name &, Entity_Class );
      virtual ~Grouping_Entity() {};
      
      struct compare
         {
	 bool operator()(Entity* x, Entity* y) const;
	 };
      multiset<Entity*, compare> Elements;                     // each element is -owned-
      typedef multiset<Entity*, compare>::const_iterator ElemIter;
      list<Entity_Container*>  auxilary_search_spaces;    //auxiliary container searched during local Name LookUp
   public:
     
      void add_AuxSpace(Entity_Container* ec) { auxilary_search_spaces.push_back(ec); }
      virtual void simple_output(ostream &os, int indent) const;
      virtual void deep_output(ostream &os, int indent) const;
      virtual void local_output(ostream &os, int indent) const;

      virtual Entity& add_Entity( Entity* e );
      virtual Entity* Local_Name_LookUp(const Entity_Name& en) const; // search only in the Grouping_Entity
      virtual AggregateType_Entity* Local_Name_LookUp_aggregate(const Entity_Name& en) const; // search only in this, search fail by defoult
     
      bool is_here( Entity * e )
         {return e->get_container()==(Entity_Container*)this; }
      virtual void change_name( Entity*, const Entity_Name& );
   };


class Extender_Methods_Container_Entity: public Grouping_Entity
   {
   public:
      Extender_Methods_Container_Entity():
           Grouping_Entity(NULL,Extender_Methods_Container)
           {}
      virtual void deep_output(ostream &os, int indent) const;
      void C_output(ostream &os, int indent) const;
   };


      // it contain one Entity that must be or Overloadable_Entity or AgrregateType_Entity
class Template_Entity: public Grouping_Entity
   {
   private:
      Token_Buffer_tiny buffer;
   public:
      Parameters_Entity& p;    // paraeters
      Token_Buffer_tiny& get_buffer()  { return buffer; }
      Template_Entity(Scope&);
      ~Template_Entity();
      
      Entity* Get_Entity() const { return *( --Elements.end()); }  // this return the entity that define the template
      virtual void simple_output(ostream &os, int indent) const;
      virtual Entity* Local_Name_LookUp(const Entity_Name& en) const;   // exclude AggregateType
      virtual AggregateType_Entity* Local_Name_LookUp_aggregate(const Entity_Name& en) const; // only AggregateType
   };




// An Overloadable_Entity is generated when is declared a function, a member or
// an operator.
// Only one between 'name' (hinerited by Entity) and 'op' has a meaning
// if  name==NULL the instance is for an operator declaration  and 'op' has meaning
//   Overloadable_Entity without name can't exist.

class Particular_overloadable;
class Overloadable_Entity: public Entity_Container
   {
   friend Grouping_Entity;
   friend Extender_Methods_Container_Entity;
   protected:
      list<Particular_overloadable*> functions;  // these are type types of the functions, elements of the list are delete-owned
      SCSPEC_value TheSC;
   public:
      Overloadable_Entity(const char *n, SCSPEC_value sc =None_SCSPEC):
         Entity_Container(n,Overloadable),
         TheSC(sc)
         {}
      Overloadable_Entity(Operator_ID p, SCSPEC_value sc =None_SCSPEC):
         TheSC(sc),
         Entity_Container(NULL,Overloadable)
         {}
      Overloadable_Entity(Entity_Name &p, SCSPEC_value sc =None_SCSPEC):
         TheSC(sc),
         Entity_Container(p,Overloadable)
         {}
      SCSPEC_value Get_SC() const 
         { return TheSC; }
      virtual void simple_output(ostream &os, int indent) const;
      virtual void deep_output(ostream &os, int indent) const;
      Particular_overloadable& add_Function( Function * t );
      const list<Particular_overloadable*>& Get_Functions() const {return functions;}
   };






// name != NULL (compulsary)
// an EnumElement_Entity can be contained only in an Enumeration_Entity
class EnumElement_Entity: public Entity
   {
   public:
      EnumElement_Entity(const char *n);
      virtual void simple_output(ostream &os, int indent) const;
   };


class Enumeration_Entity: public Grouping_Entity
   {
   public:
      Enumeration_Entity(const char *n):
         Grouping_Entity(n, Enumeration)
         {};
      virtual void simple_output(ostream &os, int indent) const;
   };








// We intend a hierarchy with founder on top and decendants on bottom
// In the private part of the following structure there are some pointers that
// describes the hierarchy. All the pointers point upward.

class AggregateType_Entity: public Grouping_Entity
   {
   private:
      list<AggregateType_Entity*> inheritance;  // the classical static inheritance
      typedef list<AggregateType_Entity*>::const_iterator InerIter;
   
   
      bool extensible;
     
                                           // only one of the following can be non NULL
      AggregateType_Entity* extend;   // class of which is an M-extender ('extend' clause) (optionally NULL) 
      AggregateType_Entity* dynamic;  // class of which is an S-extender ('dynamic' clause) (optionally NULL)
      
   public:
      Extender_Methods_Container_Entity Extender_Methods;
      void set_container( Entity_Container* ec);         // WARNING: capire se su questa funzione si fa binding statico o dinamico
      const list<AggregateType_Entity*>& Get_ineritance() const {return inheritance;};
      AggregateType_Entity(const char *n);
      AggregateType_Entity(const Entity_Name &en);
      void set_extensible();
   
      bool is_extensible() {return extensible;}
      AggregateType_Entity * Get_Extend() {return extend; }
      AggregateType_Entity * Get_Dynamic() {return dynamic;  }
      
   
      void add_inherit( AggregateType_Entity* );
      bool add_extend(AggregateType_Entity*);
      bool add_dynamic(AggregateType_Entity*);
      virtual Entity*               Scope_Name_LookUp(const Entity_Name& en) const; // search in the container, too.
      virtual AggregateType_Entity* Scope_Name_LookUp_aggregate(const Entity_Name& en) const; //
     
                  // search in this and in the parents (ascending)
      Entity*               Local_and_Inherited_Name_LookUp(const Entity_Name& en) const; // search only every entity
      AggregateType_Entity* Local_and_Inherited_Name_LookUp_aggregate(const Entity_Name& en) const; // serach only for AggregateType_Entity
           

      virtual void simple_output(ostream &os, int indent) const;
      virtual void deep_output(ostream &os, int indent) const;
   };


class Namespace_Entity: public Grouping_Entity
   {
   public:
      Namespace_Entity(const char *n):
         Grouping_Entity(n, AggregateType)
         {};
      virtual void simple_output(ostream &os, int indent) const;
   };

class Parameters_Entity: public Entity_Container
   {
   private:
   public:
      list<Entity*> Elements;
      typedef list<Entity*>::const_iterator ElemIter;
   
      Parameters_Entity():
         Entity_Container(NULL, Parameters)
         {};
      virtual void one_line_local_output(ostream &os) const;
      virtual void simple_output(ostream &os, int indent) const {};
      virtual Entity& add_Entity( Entity* e );
   };

class Function_Entity: public Entity_Container
   {
   private:
      Namespace_Entity* local;
      SCSPEC_value TheSC;              // it is not used for now (sc==None_SCSPEC)
   public:
      Function_Entity( Entity_Name & en, Function* f, SCSPEC_value sc =None_SCSPEC ):
         Entity_Container(en, Function_E),
         TheSC(sc),
         T(f),
         local( new Namespace_Entity(NULL) )
         { local->set_container(this); }
      Function *const T;    // owned
      Namespace_Entity& Get_local() const {return *local;}
      virtual void simple_output(ostream &os, int indent) const;
      virtual void deep_output(ostream &os, int indent) const;
      SCSPEC_value Get_SC() const 
         { return TheSC; }
   };



// this can be contained only in Overloadable_Entity
class Particular_overloadable: public Entity_Container
   {
   friend Overloadable_Entity;
   private:
      Particular_overloadable( Function* f, Overloadable_Entity* oe):
         T(f),
         local( new Namespace_Entity(NULL) ),
         Entity_Container( NULL,  Particular_overloadable_E )
         { 
	 local->set_container(this);
	 set_container(oe); 
	 }
      Namespace_Entity* local;
   public:
      Function *const T;
      Namespace_Entity& Get_local() const {return *local;}
   
      virtual Entity*               Local_Name_LookUp(const Entity_Name& en) const;
      virtual AggregateType_Entity* Local_Name_LookUp_aggregate(const Entity_Name& en) const;
   

      virtual void simple_output(ostream &os, int indent) const;
      virtual void deep_output(ostream &os, int indent) const;
   };


// -------------------------------------------------- Scope

class Symbol_set        // tipicamente un solo symbol set e` presente nel precompilatore
   {
   private:
      struct compare
         {
	 bool operator()(const char* x, const char* y) const;
	 };
      typedef set<const char *, compare> Container;
      Container s;
      
   public:
      const char* Update(const char* a);    // is there? --> return a pointer,
                                            // isn't there? --> make a copy and return a pointer to it
      void output(ostream& os=cerr) const;
      void gdb_out() { output(cerr); }
   };



class Scope
   {
   private:
      Entity_Container* container;      // not owned (just a reference)      
      list<Entity_Container*> stack;
   public:
      static Symbol_set symbol_table;
      Scope(Entity_Container& db):
         container(&db) 
         {}
      void Symbol_LookUp(const char* p, const char*& s)
         { s= symbol_table.Update(p); }
      Entity* Name_LookUp( const Entity_Name& en)
         { return container->Scope_Name_LookUp(en); }
      AggregateType_Entity* Name_LookUp_aggregate( const Entity_Name& en)
         { return container->Scope_Name_LookUp_aggregate(en); }
   
      void Change( Entity_Container& c)
         { container= &c; }
      Entity_Container* Get_Container() const   // return null if it hasn't a container
         { return container;}
      void Push(Entity_Container& c)
         { 
	 stack.push_front(container); 
	 container= &c;
	 }
      void Pop()
         { 
	 container= *stack.begin();
	 stack.pop_front(); 
	 }
      bool can_be_shadowed( Entity * e )
         { return !container->is_here(e); }
   
      void output(ostream& os) {container->scope_output(os, 0);}
		  
      Entity& add_Entity( Entity * e)
         { return container->add_Entity(e); }

      void cosa_vedo() 
         { 
	 output(cerr); 
	 }

   };



// Entity* Name_LookUp(char* p, Entity*& e, char*& s)  
//
// it has two return value:               
// `s' is the string in the `symbol_table'
//      this pointer i unique, is an identifier of a symbol
// `e' is NULL if no entity is reachable for that symbol in the scope
//     is the pointer to the Entity visible in the scope


// ==============================================================
const char* Get_Operator_Name(Operator_ID op);



#endif
