#ifndef FRONTEND_H
#define FRONTEND_H


// -------------------------------------------------------
// frontend.h:
// by Maurizio Pizzonia
//
// $Id: frontend.h,v 0.2 1996/08/31 16:59:56 maurizio Exp maurizio $



// 'token_string':
// in the comments in this file this word is used to denote the characters of
// the file that generate the token
class Token_Source;
class Cxx_lexer;
class Cxx_parser;

#include <fstream.h>
#include <iostream.h>

#include <bool.h>   //stl
#include <deque.h>

#include <FlexLexer.h>  //flex support to C++

#include "entity.h"
#include "pardefs.h"
#include "global.h"

#include "parser_types.h"

#include "Window_on_a_Stream.h"
#include "backend.h"



// ------------------------------------------------------- Cxx_lexer
//
//


class Cxx_lexer: private yyFlexLexer, public Token_Source 
    {
    private:
       Cxx_lexer( const Cxx_lexer & x );   // interdict copy
       Cxx_lexer& operator= (Cxx_lexer&);  // interdict assignment

       streampos stream_position;
       streampos beginning_of_pre_string;
       bool last_emitted;        // true -> last catched string was emitted as a token
                                 // false -> last catched string wasn't emitted, it was a separator
      
  
//       Window_on_a_Stream<A_token,10, Cxx_lexer> w; // this manage a window on the token stream
       

       void Update_stream_position();
       void Update_and_not_emit();
       void almost_compute_token( A_token& token, t_id t);

                                 // include file reference managment (generated by CPP preprocessor) 
       static Symbol_set files_included;
       const char* current_include; // file from which come the current line
       unsigned    current_line;    // line inside the include file from which come the current line

    public:
                     // the following function I would had like that it was
                     // private but I can't use it in the Window_on_a_Stream
                     // and then I had to put the window outside this class 
                     // (c++ (or g++?) limitations in template using)
                     // and also the access must be public.
       virtual A_token inner_lexer();   //this is created by "flex" scanner generator 
       
       Cxx_lexer(istream *in, Scope &sc, int q);
       ~Cxx_lexer() {};
    };



// ---------------------------------------------------------- Cxx_parser


class A_typed_declspecs_stack
             {          // at least one element in the list, prepare before parsing
             private:
                  list<A_typed_declspecs> stack;        
             public:
                  push() { stack.push_front(A_typed_declspecs());(*stack.begin()).null(); }
                  pull() { stack.pop_front(); }
                  A_typed_declspecs& top() { return *stack.begin(); }
             };


class Cxx_parser
   {
   private:
   
                                                        // this is not possible when we are parsing template during instantiation

  
  
  
         // the following private members are required by bison generated code
         // with %pure_parser option
      int  yylex (YYSTYPE *lvalp);
      void yyerror( char *s) { syntax_error( s );}
      int yydebug;
      int yyparse();



         // -------------- 
      Scope current_scope;
      Grouping_Entity* Global;
      Substituter   S;

                // -------------- token source managment
      Token_Source *L;            // current source of token
      Cxx_lexer *Lexer_Analizer;  // owned:  e` una istanza dell'analizzatore lessicale Cxx_lexer
           // the following intrface must be user to modify L
      list<Token_Source *> saved_L;
      void push_source( Token_Source& );  // put a source in L saving the old value
      void pop_source();                  // restore the last value of L
      bool can_modify() { return L==Lexer_Analizer; }   // this is true when it is possible to modify the source
   



      Token_Buffer* current_skip_buffer;   // this point to the buffer for saving skipped tokens
      Null_Token_Buffer Null_skip_buffer; // this is a buffer that doesn't store what you add in

  
  
  
          // ------------- syntax errors
      int count_syntax_errors;
      void syntax_error(const char* ); // called on syntax error by yyerror()
      
 // auxiliary procedures and datas
      Type* convert_expr_or_declarator(const A_name_entity_type&) const;
      Type* get_type_from_entity( Entity* ) const;    // from an entity create the relative new Type
      Type* apply_function( Type*, const list<Type*>* ) const;
      Type* applay_operator(Type*, A_token ) const;
      Entity* compute_entity( const AggregateType_Entity* E1, const A_name_entity& E2 ) const;
      AggregateType_Entity* compute_dynamic_chain(  AggregateType_Entity* descendant, list<AggregateType_Entity*>* &path ) const;

      AggregateType_Entity* get_current_class() const; // return pointer to the current class (the first starting from current scope and going outer) or NULL if it doesn't exist
      void delete_ListTypes( list<Type*>* ) const;
      void declare( A_name_entity_type& );
      void declare( A_name_entity_scope_type& );

      void template_begin_instantiation( Entity* &e, const Entity_Name& e_name );


      Function* begin_parameters();
      void end_parameters();
      void reset_type_qualifier_scope();  // if scope_for_scope_qualifier==true -> pop the scope
      bool scope_for_scope_qualifier;   // true if on the top of current_scope there is a scope for a use in qualified names
                                        // in this case it must be popped as soon as possible
  
  
      // global variables because the grammar as inherited attribute
      A_typed_declspecs_stack       current_definition;   // info for current definition used in datadef-->initdecls -->...?


      
 // these are methods for modifing the source
      void emit_CPP_info_for_generated_code(ostream& os);
      void emit_CPP_info_for_existant_code(ostream& os, A_token t);
      void freeze( streampos );
      void delete_source( streampos, streampos );
      void modify_class( AggregateType_Entity*, 
			 A_token leftcurly, 
			 A_token rightcurly );
      void modify_function( Entity*, 
			    A_token leftcurly, 
			    A_token rightcurly, 
			    bool other_initialization);
      Type* modify_access( const A_object& , const A_name_entity_scope& );        // modify an access to a member (changes occours only for dynamic inheritance)
      Type* modify_access( const A_object& , const A_name_entity& );        // modify an access to a member (changes occours only for dynamic inheritance)
      void out_extender_table_chain( char* &left,
			             char* &right,
			             list<AggregateType_Entity*>::iterator from,
			             list<AggregateType_Entity*>::iterator to) const;


// interface

   public:
      Cxx_parser( istream& in, Grouping_Entity& db, istream& in_for_emit,  int d);
      ~Cxx_parser();
      int parse();
      int how_many_syntax_errors() { return count_syntax_errors; }
   };








#endif

