


#include <iostream.h>
#include <stdlib.h>
#include "global.h"
#include "messages.h"
#include "errno.h"


int M::count_errors=0;
int M::count_warnings=0;

int M::count_internal_warnings=0;
int M::count_internal_errors=0;

int M::warning_level=5;
int M::count_non_displayed_warnings=0;

bool M::display_preprocessed_line_number=false;
bool M::display_internal_warning=false;


M::Warning  M::Already_defined= {8, "symbol already defined"};
M::Warning  M::Type_Mismatch=    {8, "type mismatch in expression"};

#define EFatal M::fatal
#define ERecov M::rec

M::Error  M::Output_Error=     {EFatal, "output error"};
M::Error  M::Input_Error=      {EFatal, "input error"};
M::Error  M::Non_ECO_source=    {EFatal, "Non ECO C++ source or \"ECO.h\" not include at first place"};
M::Error  M::Too_Complex=      {ERecov, "I can't recognize type before .{ or -{, expression too complex"};
M::Error  M::First_Costr_Parm= {ERecov, "constructor must have a named 1st parameter pointer or reference to M-extended class or top of a dynamic hierarchy" };
M::Error  M::Template_Wrong_ParNum= {ERecov, "template used with wrong number of parameters"};
M::Error  M::Wrong_overloading= {ERecov, "wrong overloading"};
M::Error  M::Undeclared_symbol= {ERecov, "undeclared symbol"};
M::Error  M::Wrong_member=     {ERecov, "wrong member access"};

#define IFatal M::fatal
#define IRecov M::rec  
#define IWarn  M::warn 

M::Internal M::Mult_E_or_D=  {IRecov, "AggregateType_Entity::add_extend(AggregateType_Entity*) multiple 'extend' or 'dynamic' added"};

M::Internal M::Past_The_End=   {IFatal, "Substituter::write_seq(...): read past the end of input"};
M::Internal M::Useless_Freeze= {IWarn,  "Substituter::freeze(streampos till_here), useless freeze()"};
M::Internal M::Freeze_in_Subst={IFatal, "Substituter::freeze(...): freeze in the middle of a substitution."};
M::Internal M::Subst_Exist=    {IFatal, "Substituter::substitute(): a substitution for this value of 'where' already exist"};
M::Internal M::Subst_Overlap=  {IFatal, "Substituter::substitute(): substitution overlap"};
M::Internal M::Subst_NULL=     {IFatal, "Substituter::substitute(): s==NULL"};
M::Internal M::Already_Freezed={IFatal, "Substituter::substitute(): can't susbstitute already freezed part"};

M::Internal M::Token_filter=   {IFatal, "Token_Source::filter(): unexpected token"};
M::Internal M::Bad_request=    {IRecov, "Cxx_parser::declare(): bad declaration request"};


  
  
     void
M::
message_header(const char* t)
     { 
     cerr << "eco_prec ";
       if ( display_preprocessed_line_number )
         cerr << filename << " " << Line;
       else 
	 cerr << include_file << " " << include_file_line;
     cerr << " " << t << ": ";
     }





     void 
M::
warning( unsigned level,  const char * s)
     {
     if ( level <= warning_level )
        {
        message_header( "warning" );
        cerr << "(" << level<< ") " << s << endl;
        ++ count_warnings;
	}
     else
        ++ count_non_displayed_warnings;
     };



     void 
M::
internal(const char*s)
     {
     hl_internal(fatal, s);
     }




     void 
M::
winternal(const char*s) 
     {
     hl_internal(warn, s);
     }




     void 
M::
out_statistics( ostream& os ) 
     {
     os << "Errors: "            << count_errors << endl
        << "Warnings: "          << count_warnings << endl
        << "Non-displyed warnings: " << count_non_displayed_warnings << endl
        << endl
        << "Internal warnings: " << count_internal_warnings 
        << " ("<< (display_internal_warning? "displayed" : "not displayed") << ")" << endl
        << "Internal errors: " << count_internal_errors << endl;
     }





     void 
M::
hl_error(Error& e) 
     {
     if ( e.t==fatal )
         message_header("Fatal error");
     else
         message_header( "Error" );
     
     cerr << e.s;
     
     if ( &e == &Output_Error ||  &e == &Input_Error )
            cerr << ": " << strerror(errno);
     cerr << endl;
     ++count_errors;
     if (e.t==fatal) throw "fatal error";
     }




     void 
M::
hl_warning(Warning& w)
     {
     if ( w.level <= warning_level )
          {
	  message_header("Warning");
          cerr << w.s << endl;
          ++ count_warnings;
	  }
     else
          ++ count_non_displayed_warnings;
     }



     void 
M::
hl_internal(Internal& e)
     {
     hl_internal(e.t, e.s);
     }



     void 
M::
hl_internal(M::type t, const char* s)
     {
     char* mssg=NULL;
     bool display=true;
     switch(t)
        {
	case fatal: mssg="Fatal internal error"; break;
	case rec  : mssg="Internal error"; break;
	case warn : 
	   mssg="Internal warning"; 
	   display= display_internal_warning;
	   break;
	}
     
     if(display)
        {
        message_header( mssg );
        cerr << s << endl;
        }
     
     switch(t)
        {
	case fatal: throw "Fatal internal error"; break;
	case rec  : ++count_internal_errors; break;
	case warn : ++count_internal_warnings; break;
	}
     }


