
#ifndef __PARSER_types_h__
#define __PARSER_types_h__

#include "global.h"
// The followin structure are usefoul when a grammar symbol must return 
// a set of values.
//
// naming conventions:
// the name of the struct is obtained from the grammar symbol preceded by 'A_'
// the corrisponding union YYSTYPE field is obtained by it preceded bya 'U'
//
// example:
// grammar symbol--> declmod
// struct name   --> A_declmod
// YYSTYPE field --> UA_declmod


class Type;


struct common
    {
    streampos begin;    // positions where the matched string begin and end
    streampos end;      // a value of zero mean "unspecified"
    };
    
    
struct A_declmods
    {
    SCSPEC_value     sc;   // storage class
    TYPE_QUAL_value  tq;   // type qualifiers
    void null() { sc=None_SCSPEC; tq=None_TYPE_QUAL; }
    };
struct A_typed_declspecs
    {
    SCSPEC_value     sc;   // storage class
    TYPE_QUAL_value  tq;   // type qualifiers
    Type*            type; // the type      
    void null() { sc=None_SCSPEC; tq=None_TYPE_QUAL; type=NULL; }
    };

struct A_typespecqual_reserved
    {
    Type* type;          // solo uno e` significativo: o tq==None_TYPE_QUAL o type==NULL
    TYPE_QUAL_value tq;  
    };
struct A_identifier: public common
    {
    const char *s;          // `s' e` sempre significativo
    Entity* e;              // `e' e` significativo solo se l'identificatore era gia` definito
    };

struct A_Operator_Identifier: public common
    {
    Operator_Identifier*   oi;
    };


struct A_name_entity: public common
    {
    Entity_Name* name;
    Entity*      e;
    };


struct A_EntityContainer: public common
    {
    Entity_Container*  c;
    };

struct A_type: public common
    {
    Type*  t;
    };


struct A_name_entity_scope: public common
    {
    Entity_Name*     name;
    Entity*          e;        //  not always meaningful
    Entity_Container* scope;    //  always meaningful (and redundant when `e' is meaningful)
    };

struct A_name_entity_type
    {
    Entity_Name* name;
    Entity*      e;
    Type*        t;
    };

struct A_name_entity_scope_type
    {
    Entity_Name* name;
    Entity*      e;            //  not always meaningful
    Entity_Container* scope;    //  always meaningful (and redundant when `e' is meaningful)
    Type*        t;
    };

struct A_Entity_endstring
    {
    Entity* e;
    streampos end_string;
    operator Entity* () {return e;}
    };

struct A_Entity_Name_endstring
    {
    Entity_Name* n;
    streampos end_string;
    operator Entity_Name* () {return n;}
    };


struct A_expression: public common
    {
    Type *t;      // type of the subexpression, it must be deleted
    streampos begin, end;  
    };



// the following is used only as attribute of "object" which reduce the following forms
//  <primary> .                                  (point)
//  <primary> ->                                 (pointsat)
//  <primary> .{ <complete_type_name> }.         (point_dynamic)
//  <primary> -{ <complete_type_name> }>         (pointsat_dynamic)
//  <primary> .{ <complete_type_name> }          (get_pointer)
struct A_object:public common
    {
    enum { get_pointer, point, pointsat, point_dynamic, pointsat_dynamic } k;
    A_expression   primary;
    Entity*        e;        // it has meaning only for k== point_dynamic or k== pointsat_dynamic
      
    streampos begin_oper, end_oper;  // begin/end of ".{xxx}." or "-{xxx}>" or "." or "->"
    };



// =====================================================================
struct value_identifier           
        {
        const char *s;          // `s' e` sempre significativo
        Entity* e;              // `e' e` significativo solo se l'identificatore era gia` definito
        };


union Value
    { 
    Operator_ID op;    // operator
    SCSPEC_value  sc;  // storage class
    VISSPEC_value  vs; // visibility specifier
    TYPESPEC_value ts; // type specifier
    TYPE_QUAL_value tq; // type qulifier
             
    value_identifier id; // identifier
    };

class token;
struct  A_token   
    {
    t_id tokenid;
                                                                     // position in the CPP output
    unsigned line;                // line in which is the token
    streampos begin_pre_string;   // first char of the pre_string
    streampos begin_token_string; // first char of the token_string
    streampos end_token_string;   // last char of the token_string
                                                                      // position in the CPP input
    const char* include_file;     // file from which the token comes
    unsigned    include_file_line;// line of the file above
    
    Value aux;   // an auxiliar value for the identification of token (actualy only operator)  related to multiple string
    void init_from_token(token*);
    const char* token_string() const;
    A_token* operator ->() {return this;}
    };
ostream& operator<< (ostream& os, const A_token& t);




typedef A_token A_token_SCSPEC;
typedef A_token A_token_VISSPEC;
typedef A_token A_token_TYPESPEC;
typedef A_token A_token_TYPE_QUAL;
typedef A_token A_token_IDENTIFIER;
typedef A_token A_token_TYPENAME;



#endif
