#ifndef __TOKEN_H__
#define __TOKEN_H__


#include <iostream.h>

#include "global.h"
#include "entity.h"

class token;
#include "pardefs.h"

// ====================================================================


class token
   {
   private:
 //this could be declared constant
     t_id tokenid;
   
   public:
   
 //these could be declared const
                   // these allow to recover, from the input stream, 
                   // the text associated to the token
     streampos begin_pre_string;   // first char of the pre_string
     streampos begin_token_string; // first char of the token_string
     streampos end_token_string;   // last char of the token_string
               // properties: 
               // tokens are orderly generated like the order in the input stream
               // begin_pre_string <= begin_token_string < end_token_string
               //        first "less then" is non-strict: a token can have no pre_string
               //        amost always second "less then" is strict: a token_stirng has lenght>0 unless is something strange like END_OF_FILE or something artificially added
                                
               // given two consecutive tokens t1, t2 we have:
               //       t1.end_token_string+1 == t2.begin_token_string

     
     
     
     static istream *input_stream;   // input stream used by "emit(.)" to take real characters for the token;
     void emit(ostream& os) const;   // take string from begin_pre_string to end_token_string from input_stream 
                                     // and output it on 'os'
     
     
     
     
     
     Operator_ID aux;   // an auxiliar value for the identification of token (actualy only operator)  related to multiple string
                        // with this some of the other token value could be considered redoundant but
                        // maybe the others are more important than operators like == != or <= >=
     
     token(t_id t,
	   streampos begin_pre, 
	   streampos begin_token, 
	   streampos end_token,
	   Operator_ID auxiliar=Op_None
 	   ):
        tokenid(t),
        begin_pre_string(begin_pre),
        begin_token_string(begin_token),
        end_token_string(end_token),
        aux(auxiliar)
        {};
     token()
        {}
   /*
     token& operator=( const token& t )
        {
	tokenid=t.tokenid;     
        begin_pre_string=t.begin_pre_string;  
        begin_token_string=t.begin_token_string;
        end_token_string=t.end_token_string;  
	}
   */

     virtual ~token() {}                    

   
     t_id getid() const
        { return tokenid; }
   /*
     void setid( t_id x ) 
        { tokenid= x; }
    */
   
     virtual const char * token_string() const
        { return "unknown";}
   };

class token_SCSPEC: public token
   {
   public:
      SCSPEC_value which;
      token_SCSPEC( SCSPEC_value v, 
		    streampos begin_pre, 
                    streampos begin_token, 
	            streampos end_token ):
          token(SCSPEC , begin_pre, begin_token, end_token),
          which(v)
          {}
   };


class token_VISSPEC: public token
   {
   public:
      VISSPEC_value which;
      token_VISSPEC( VISSPEC_value v, 
		    streampos begin_pre, 
                    streampos begin_token, 
	            streampos end_token ):
          token(VISSPEC, begin_pre, begin_token, end_token),
          which(v)
          {}
   };


class token_TYPESPEC: public token
   {
   public:
      TYPESPEC_value which;
      token_TYPESPEC( TYPESPEC_value v, 
		    streampos begin_pre, 
                    streampos begin_token, 
	            streampos end_token ):
          token(TYPESPEC , begin_pre, begin_token, end_token),
          which(v)
          {}
   };

class token_TYPE_QUAL: public token
   {
   public:
      TYPE_QUAL_value which;
      token_TYPE_QUAL( TYPE_QUAL_value v, 
		    streampos begin_pre, 
                    streampos begin_token, 
	            streampos end_token ):
          token(TYPE_QUAL , begin_pre, begin_token, end_token),
          which(v)
          {}
   };

class token_IDENTIFIER: public token
   {
   public:
      const char *s;
      Entity * e;     // if e!=NULL then `s' is redoundant.
      token_IDENTIFIER( const char* v, 
		    streampos begin_pre, 
                    streampos begin_token, 
	            streampos end_token ):
          token(IDENTIFIER , begin_pre, begin_token, end_token),
          s(v),
          e(NULL)
          {}
      token_IDENTIFIER( Entity* v, 
		    streampos begin_pre, 
                    streampos begin_token, 
	            streampos end_token ):
          token(IDENTIFIER , begin_pre, begin_token, end_token),
          e(v),
          s(v->name)
          {}
       virtual const char * token_string() const
          { return s;}
   };


class token_TYPENAME: public token
   {
   public:
      const char *s;
      Entity * e;     // it is alaways != NULL, see next_token() and the way to create token_TYPENAME from token_IDENTIFIER
      token_TYPENAME( const token_IDENTIFIER& t):
          token(TYPENAME , t.begin_pre_string, t.begin_token_string, t.end_token_string),
          s(t.s),
          e(t.e)
          {}
       virtual const char * token_string() const
          { return s;}
   };
 

    
    
   
inline ostream& operator<< (ostream& os, const token& t)
   {
   os   << "t_id=" <<t.getid() << "   "
        << t.token_string() << "   "
        << t.begin_pre_string << "   "
        << t.begin_token_string << "   "
        << t.end_token_string ;   
   return os;
   };

#endif
